#!/usr/bin/env python

"""buildpkg.py -- Build OS X packages for Apple's Installer.app.

This is an experimental command-line tool for building packages to be
installed with the Mac OS X Installer.app application. 

Please read the file ReadMe.txt for more information!

Dinu C. Gherman, 
gherman@europemail.com
September 2002

!! USE AT YOUR OWN RISK !!
"""

__version__ = 0.3
__license__ = "FreeBSD"


import os, sys, glob, fnmatch, shutil, string, copy, getopt
from os.path import basename, dirname, join, islink, isdir, isfile

Error = "buildpkg.Error"

PKG_INFO_FIELDS = """\
Title
Version
Description
DefaultLocation
Diskname
DeleteWarning
NeedsAuthorization
DisableStop
UseUserMask
Application
Relocatable
Required
InstallOnly
RequiresReboot
InstallFat\
"""

######################################################################
# Helpers
######################################################################

# Convenience class, as suggested by /F.

class GlobDirectoryWalker:
    "A forward iterator that traverses files in a directory tree."

    def __init__(self, directory, pattern="*"):
        self.stack = [directory]
        self.pattern = pattern
        self.files = []
        self.index = 0


    def __getitem__(self, index):
        while 1:
            try:
                file = self.files[self.index]
                self.index = self.index + 1
            except IndexError:
                # pop next directory from stack
                self.directory = self.stack.pop()
                self.files = os.listdir(self.directory)
                self.index = 0
            else:
                # got a filename
                fullname = join(self.directory, file)
                if isdir(fullname) and not islink(fullname):
                    self.stack.append(fullname)
                if fnmatch.fnmatch(file, self.pattern):
                    return fullname


######################################################################
# The real thing
######################################################################

class PackageMaker:
    """A class to generate packages for Mac OS X.

    This is intended to create OS X packages (with extension .pkg)
    containing archives of arbitrary files that the Installer.app 
    (Apple's OS X installer) will be able to handle.

    As of now, PackageMaker instances need to be created with the 
    title, version and description of the package to be built. 
    
    The package is built after calling the instance method 
    build(root, resources, **options). The generated package is 
    a folder hierarchy with the top-level folder name equal to the 
    constructor's title argument plus a '.pkg' extension. This final
    package is stored in the current folder.
    
    The sources from the root folder will be stored in the package
    as a compressed archive, while all files and folders from the
    resources folder will be added to the package as they are.

    Example:
    
    With /my/space being the current directory, the following will
    create /my/space/distutils-1.0.2.pkg/:

      PM = PackageMaker
      pm = PM("distutils-1.0.2", "1.0.2", "Python distutils.")
      pm.build("/my/space/sources/distutils-1.0.2")
      
    After a package is built you can still add further individual
    resource files or folders to its Contents/Resources subfolder
    by using the addResource(path) method: 

      pm.addResource("/my/space/metainfo/distutils/")
    """

    packageInfoDefaults = {
        'Title': None,
        'Version': None,
        'Description': '',
        'DefaultLocation': '/',
        'Diskname': '(null)',
        'DeleteWarning': '',
        'NeedsAuthorization': 'NO',
        'DisableStop': 'NO',
        'UseUserMask': 'YES',
        'Application': 'NO',
        'Relocatable': 'YES',
        'Required': 'NO',
        'InstallOnly': 'NO',
        'RequiresReboot': 'NO',
        'InstallFat': 'NO'}


    def __init__(self, title, version, desc):
        "Init. with mandatory title/version/description arguments."

        info = {"Title": title, "Version": version, "Description": desc}
        self.packageInfo = copy.deepcopy(self.packageInfoDefaults)
        self.packageInfo.update(info)
        
        # variables set later
	self.packageContentsFolder = None
        self.packageRootFolder = None
        self.packageResourceFolder = None
        self.sourceFolder = None
        self.resourceFolder = None


    def _escapeBlanks(self, s):
        "Return a string with escaped blanks."
        
        return s.replace(' ', '\ ')
                

    def build(self, root, resources=None, **options):
        """Create a package for some given root folder.

        With no 'resources' argument set it is assumed to be the same 
        as the root directory. Option items replace the default ones 
        in the package info.
        """

        # set folder attributes
        self.sourceFolder = root

        if resources == None:
            self.resourceFolder = None
        else:
            self.resourceFolder = resources

        # replace default option settings with user ones if provided
        fields = self. packageInfoDefaults.keys()
        for k, v in options.items():
            if k in fields:
                self.packageInfo[k] = v
            elif not k in ["OutputDir"]:
                raise Error, "Unknown package option: %s" % k
        
        # Check where we should leave the output. Default is current directory
        outputdir = options.get("OutputDir", os.getcwd())
        packageName = self.packageInfo["Title"]
        self.packageRootFolder = os.path.join(outputdir, packageName + ".pkg")
 
        # do what needs to be done
        self._makeFolders()
	self._addInfo()
	self._addArchive()
        self._addBom()
        self._addResources()
        self._addSizes()


    def addResource(self, path):
        "Add arbitrary file or folder to the package resource folder."
        
        # Folder basenames become subfolders of Contents/Resources.
        # This method is made public for those who wknow what they do!
   
        prf = self.packageResourceFolder
        if isfile(path) and not isdir(path):
            shutil.copy(path, prf)
        elif isdir(path):
            path = self._escapeBlanks(path)
            prf = self._escapeBlanks(prf)
            os.system("cp -r %s %s" % (path, prf))
        

    def _makeFolders(self):
        "Create package folder structure."

        # Not sure if the package name should contain the version or not...
        # packageName = "%s-%s" % (self.packageInfo["Title"], 
        #                          self.packageInfo["Version"]) # ??

        self.packageContentsFolder = join(self.packageRootFolder, "Contents")
        self.packageResourceFolder = join(self.packageContentsFolder, "Resources")
        os.mkdir(self.packageRootFolder)
        os.mkdir(self.packageContentsFolder)
        os.mkdir(self.packageResourceFolder)


    def _addInfo(self):
        "Write .info file containing installing options."

        # Not sure if options in PKG_INFO_FIELDS are complete...

        info = ""
        for f in string.split(PKG_INFO_FIELDS, "\n"):
            info = info + "%s %%(%s)s\n" % (f, f)
        info = info % self.packageInfo
        base = self.packageInfo["Title"] + ".info"
        path = join(self.packageResourceFolder, base)
        f = open(path, "w")
        f.write(info)


    def _addBom(self):
        "Write .bom file containing 'Bill of Materials'."

        # Currently ignores if the 'mkbom' tool is not available.

        try:
            base = self.packageInfo["Title"] + ".bom"
            bomPath = join(self.packageResourceFolder, base)
            bomPath = self._escapeBlanks(bomPath)
            sourceFolder = self._escapeBlanks(self.sourceFolder)
            cmd = "mkbom %s %s" % (sourceFolder, bomPath)
            res = os.system(cmd)
        except Exception:
            pass


    def _addArchive(self):
        "Write .pax.gz file, a compressed archive using pax/gzip."


# FIX SYMBOLIC
	cwd = os.getcwd()
	os.chdir(self.sourceFolder)
        # Pax is the enemy.  Make war.
	sitepackages = join("System", "Library", "Frameworks", "Python.framework", "Versions", "2.3", "lib", "python2.3", "site-packages")

	if isdir(join(cwd, self.sourceFolder, sitepackages)):
	   librarypython23 = join("Library", "Python", "2.3")
	   if not isdir(librarypython23):
	   	os.makedirs(librarypython23)
	   shutil.move(join(cwd, self.sourceFolder, sitepackages), join(cwd, self.sourceFolder, librarypython23))

# END FIXING OF THE SYMBOLIC LINKS

        # Currently ignores if the 'pax' tool is not available.

        # create archive
        base = basename(self.packageInfo["Title"]) + ".pax"
        self.archPath = join(self.packageResourceFolder, base)
	working = join(cwd, self.archPath)
        archPath = self._escapeBlanks(working)
        cmd = "pax -wf %s %s" % (archPath, ".")
	print archPath
        res = os.system(cmd)
        
        # compress archive
        cmd = "gzip %s" % archPath
        res = os.system(cmd)
        os.chdir(cwd)


    def _addResources(self):
        "Add all files and folders inside a resources folder to the package."

        # This folder normally contains Welcome/ReadMe/License files, 
        # .lproj folders and scripts.

        if not self.resourceFolder:
            return

        files = glob.glob("%s/*" % self.resourceFolder)
        for f in files:
            self.addResource(f)
        

    def _addSizes(self):
        "Write .sizes file with info about number and size of files."

        # Not sure if this is correct, but 'installedSize' and 
        # 'zippedSize' are now in Bytes. Maybe blocks are needed? 
        # Well, Installer.app doesn't seem to care anyway, saying 
        # the installation needs 100+ MB...

        numFiles = 0
        installedSize = 0
        zippedSize = 0

        files = GlobDirectoryWalker(self.sourceFolder)
        for f in files:
            numFiles = numFiles + 1
            installedSize = installedSize + os.lstat(f)[6]

        try:
            zippedSize = os.stat(self.archPath+ ".gz")[6]
        except OSError: # ignore error 
            pass
        base = self.packageInfo["Title"] + ".sizes"
        f = open(join(self.packageResourceFolder, base), "w")
        format = "NumFiles %d\nInstalledSize %d\nCompressedSize %d\n"
        f.write(format % (numFiles, installedSize, zippedSize))

# Shortcut function interface

def buildPackage(*args, **options):
    "A shortcut function for building a package."
    
    o = options
    title, version, desc = o["Title"], o["Version"], o["Description"]
    pm = PackageMaker(title, version, desc)
    apply(pm.build, list(args), options)

    return pm


######################################################################
# Command-line interface
######################################################################

def printUsage():
    "Print usage message."

    format = "Usage: %s <opts1> [<opts2>] <root> [<resources>]"
    print format % basename(sys.argv[0])
    print
    print "       with arguments:"
    print "           (mandatory) root:         the package root folder"
    print "           (optional)  resources:    the package resources folder"
    print
    print "       and options:"
    print "           (mandatory) opts1:"
    mandatoryKeys = string.split("Title Version Description", " ")
    for k in mandatoryKeys:
        print "               --%s" % k
    print "           (optional) opts2: (with default values)"

    pmDefaults = PackageMaker.packageInfoDefaults
    optionalKeys = pmDefaults.keys()
    for k in mandatoryKeys:
        optionalKeys.remove(k)
    optionalKeys.sort()
    maxKeyLen = max(map(len, optionalKeys))
    for k in optionalKeys:
        format = "               --%%s:%s %%s"
        format = format % (" " * (maxKeyLen-len(k)))
        print format % (k, repr(pmDefaults[k]))


def main():
    "Command-line interface."

    shortOpts = ""
    keys = PackageMaker.packageInfoDefaults.keys()
    longOpts = map(lambda k: k+"=", keys)

    try:
        opts, args = getopt.getopt(sys.argv[1:], shortOpts, longOpts)
    except getopt.GetoptError, details:
        print details
        printUsage()
        return

    optsDict = {}
    for k, v in opts:
        optsDict[k[2:]] = v

    ok = optsDict.keys()
    if not (1 <= len(args) <= 2):
        print "No argument given!"
    elif not ("Title" in ok and \
              "Version" in ok and \
              "Description" in ok):
        print "Missing mandatory option!"
    else:
        pm = apply(buildPackage, args, optsDict)
        return

    printUsage()

    # sample use:
    # buildpkg.py --Title=distutils \
    #             --Version=1.0.2 \
    #             --Description="Python distutils package." \
    #             /Users/dinu/Desktop/distutils


if __name__ == "__main__":
    main()



# -*- mode: python; indent-tabs-mode: nil; tab-width: 3 -*-
# vim: set tabstop=3 shiftwidth=3 expandtab:
#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom, and Karl MacMillan
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import wx
from gamera import core, config
from gamera.gui import gamera_display, gui_util, gamera_icons
import glob, string, os.path

class FileList(wx.GenericDirCtrl):
   def __init__(self, parent, ID, image_display):
      wx.GenericDirCtrl.__init__(
          self, parent, ID,
          filter="All files (*.*)|*.*|TIFF files (*.tiff,*.tif)|*.tiff,*.tif",
          style=wx.DIRCTRL_SHOW_FILTERS)
      # self.SetDefaultPath(config.options.file.default_directory)
      wx.EVT_TREE_ITEM_ACTIVATED(self.GetTreeCtrl(), -1, self.OnItemSelected)
      self.image_display = image_display

   def OnItemSelected(self, e):
      filename = self.GetFilePath()
      if filename == '':
         return
      wx.BeginBusyCursor()
      try:
         filename = filename.encode('utf8')
      except Exception:
         pass
      try:
         try:
            image = core.load_image(filename)
         except Exception, e:
            gui_util.message("Loading image %s failed. \n\nThe error was:\n%s"
                             % (filename, str(e)))
            return
      finally:
         wx.EndBusyCursor()
      width, height = self.image_display.id.GetSize()
      scale = max(float(width) / float(image.width),
                  (float(height) / float(image.height)))
      self.image_display.id.set_image(image, weak=0)
      self.image_display.id.scale(scale)

class ImageBrowserFrame(wx.Frame):
   def __init__(self):
      wx.Frame.__init__(self, None, -1, "Image File Browser",
                       wx.DefaultPosition,(600, 400))
      icon = wx.IconFromBitmap(gamera_icons.getIconImageBrowserBitmap())
      self.SetIcon(icon)
      self.splitter = wx.SplitterWindow(self, -1)
      self.image = gamera_display.ImageWindow(self.splitter, -1)
      self.file = FileList(self.splitter, -1, self.image)
      self.splitter.SetMinimumPaneSize(20)
      self.splitter.SplitVertically(self.file, self.image)
      self.splitter.Show(1)
      self.image.id.RefreshAll()





# -*- coding: utf-8 -*-
#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom, and Karl MacMillan
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from gamera.plugin import *
import _thinning

class Thinning(PluginFunction):
    self_type = ImageType([ONEBIT])
    return_type = ImageType([ONEBIT])
    doc_examples = [(ONEBIT,)]

class thin_zs(Thinning):
    """
    Thins (skeletonizes) a ONEBIT image using the Zhang and Suen
    algorithm.

    The resulting skeleton is not a medial axis transformation, and
    the ends of the skeleton will not extend to the edges of the
    original image.

    T. Y. Zhang and C. Y. Suen. 1984.  A Fast Parallel Algorithm for
    Thinning Digital Patterns., *Communications of ACM*, 2(3).
  
    R. C. Gonzalez and P. Wintz. 1987 *Digital Image Processing.*,
    2. edition. 398-402.
    """
    pass

    
class thin_hs(Thinning):
    """
    Derives the medial axis transformation from a ONEBIT image using the
    Haralick and Shapiro algorithm.

    Unlike thin_zs_ and thin_lc_, this function performs a medial axis
    transformation, and the ends of the skeleton extend to the corners of
    the original image.

    Consider using thin_hs_large_image_ instead, for faster performance on
    large images with a lot of connected components.

    R. M. Haralick and L. G. Shapiro. 1992.  *Computer and Robot Vision*,
    Vol. 1, Chapter 5 (especially 5.10.1).  Reading, MA: Addison-Wesley.
    """
    pass

class medial_axis_transform_hs(Thinning):
    """"This function is an alias for thin_hs_."""
    pure_python = True
    def __call__(self):
        return _thinning.thin_hs(self)

class thin_hs_large_image(Thinning):
    """
    Thins (skeletonizes) a ONEBIT image using the Haralick and Shapiro
    algorithm.

    Unlike thin_hs_, this algorithm performs skeletonization on one
    connected component at a time.  On large images with a lot of
    connected components, this can be significantly faster.  However,
    for small images with a single connected component, this has
    unnecessary overhead, which is why both versions are included.
    Please note cc_analysis results in a labelled image, which you can
    reset afterwards with reset_onebit_image().
    """
    pure_python = True
    def __call__(self):
        copy = self.image_copy()
        ccs = copy.cc_analysis()
        for cc in ccs:
            thin_cc = cc.thin_hs()
            cc.and_image(thin_cc, in_place=True)
        return copy
    __call__ = staticmethod(__call__)

class medial_axis_transform_large_image_hs(Thinning):
    """
    This function is an alias for thin_hs_large_image_.
    """
    pure_python = True
    def __call__(self):
        return thin_hs_large_image()(self)

class thin_lc(Thinning):
    """
    Thins (skeletonizes) a ONEBIT image using the Lee and Chen
    algorithm.

    This function is a simple extension to the Zhang and Suen
    algorithm in thin_zs_ that ensure that no two pixels are more than
    4-connected.
    
    The resulting skeleton is not a medial axis transformation, and
    the ends of the skeleton will not extend to the edges of the
    original image.

    H.-J. Lee and B. Chen. 1992.  Recognition of handwritten Chinese
    characters via short line segments. *Pattern Recognition*. 25(5)
    543-552.
    """
    pass

class ThinningModule(PluginModule):
    category = "Thinning"
    functions = [thin_zs, thin_hs, medial_axis_transform_hs,
                 thin_hs_large_image, medial_axis_transform_large_image_hs,
                 thin_lc]
    cpp_headers = ["thinning.hpp"]
    author = u"Michael Droettboom and Karl MacMillan (based on code by \u00d8ivind Due Trier and Qian Huang)"
    url = "http://gamera.sourceforge.net/"
module = ThinningModule()

del Thinning



# -*- mode: python; indent-tabs-mode: nil; tab-width: 3 -*-
# vim: set tabstop=3 shiftwidth=3 expandtab:
#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom,
#                          and Karl MacMillan
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Utilities to make writing setup.py files for Gamera extensions easier"""

import sys
import os
import platform
import glob
from distutils.sysconfig import get_python_lib, get_python_inc, PREFIX
from distutils.command import install_data

# Fix RPM building
#
# This is a total hack to patch up an error in distutils.command.bdist_rpm
import distutils.command.bdist_rpm
def rpm_run(self):
   try:
      original_rpm_run(self)
   except AssertionError, e:
      if str(e).startswith("unexpected number of RPM files found"):
         rpms = glob.glob(os.path.join(os.path.join(self.rpm_base, 'RPMS'),
                                       "*/*.rpm"))
         for rpm in rpms:
            self.move_file(rpm, self.dist_dir)
      else:
         raise e
original_rpm_run = distutils.command.bdist_rpm.bdist_rpm.run
setattr(distutils.command.bdist_rpm.bdist_rpm, 'run', rpm_run)

class smart_install_data(install_data.install_data):
   def run(self):
      install_cmd = self.get_finalized_command("install")
      install_dir = os.path.join(getattr(install_cmd, "install_lib"), "gamera")
      print "INSTALL DIRECTORY", install_dir
      output = []
      for path, files in self.data_files:
         if "$LIB" in path:
            path = path[path.find("$LIB"):]
            path = path.replace("$LIB", install_dir)
         output.append((path, files))
      self.data_files = output
      return install_data.install_data.run(self)

cmdclass = {'install_data': smart_install_data}
if sys.platform == "darwin":
   from gamera.mac import gamera_mac_setup
   cmdclass['bdist_osx'] = gamera_mac_setup.bdist_osx
elif sys.platform == "win32":
   try:
      from win32 import bdist_msi
      cmdclass['bdist_msi'] = bdist_msi.bdist_msi
   except ImportError:
      pass

# If gamera.generate is imported gamera.__init__.py will
# also be imported, which won't work until the build is
# finished. To get around this, the gamera directory is
# added to the path and generate is imported directly
sys.path.append("gamera")
import generate

extras = {'extra_compile_args': ['-Wall']}
if sys.platform == 'win32' and not '--compiler=mingw32' in sys.argv:
   extras['extra_compile_args'] = ['/GR','/EHsc']#, "/Zi"]
elif sys.platform == 'darwin':
   extras['extra_link_args'] = ['-F/System/Library/Frameworks/']
   macosversion = platform.mac_ver()[0]
   if macosversion.startswith("10.6"):
      extras['extra_link_args'].append('-L/Developer/SDKs/MacOSX10.5.sdk/usr/lib')
elif '--compiler=mingw32' in sys.argv or not sys.platform == 'win32':
   extras['libraries'] = ['stdc++'] # Not for intel compiler

# Check that we are running a recent enough version of Python.
# This depends on the platform.
default_required_version = 222
required_versions = {'linux2': 222, 'win32':  231, 'darwin': 230, 'cygwin': 222}
version = float(''.join([str(x) for x in sys.version_info[0:3]]))
required_version = required_versions.setdefault(sys.platform, default_required_version)
if version < required_version:
   print "Gamera requires Python version %s or later." % '.'.join(list(str(required_version)))
   print "You are running the following Python version:"
   print sys.version
   sys.exit(1)

lib_path = os.path.join(get_python_lib()[len(PREFIX)+1:], "gamera")
include_path = os.path.join(get_python_inc()[len(PREFIX)+1:], "gamera")

def get_plugin_filenames(path):
   """Return all of the python plugin files in a specified path. This is not
   the same as glob.glob('*.py') in that it removes __init__.py files and
   normalizes the path in an os independent way."""
   plugins = glob.glob(path + "/*.py")
   norm_plugins = []
   for x in plugins:
      norm_plugins.append(os.path.normpath(os.path.abspath(x)))
   plugins = norm_plugins
   try:
      path = os.path.normpath(os.path.abspath(path + "/__init__.py"))
      plugins.remove(path)
   except Exception:
      pass
   return plugins

def generate_plugins(plugins, location, compiling_gamera=0):
   """Generate the necessary cpp wrappers from a list of python plugin
   filenames. The regeneration only happens if it is necessary (either
   the python file has changed or one of the files that it depends on
   has changed). A distutiles extension class is created for each plugin
   that needs to be compiled."""

   # Create the list of modules to ignore at import - because
   # we are in the middle of the build process a lot of C++
   # plugins don't yet exist. By preventing the import of
   # the core of gamera and all of the plugins we allow the
   # plugins to be imported for the build process to examine
   # them. Some of this is unnecessary for external plugins,
   # but it shouldn't hurt.
   ignore = ["core", "gamera.core", "gameracore"]
   for x in plugins:
      plug_path, filename = os.path.split(x)
      module_name = "_" + filename.split('.')[0]
      ignore.append(module_name)
   generate.magic_import_setup(ignore)
   
   plugin_extensions = []
   for x in plugins:
      extension = generate.generate_plugin(
         x, location, compiling_gamera, **extras)
      if not extension is None:
         plugin_extensions.append(extension)
   
   generate.restore_import()
   return plugin_extensions

def get_gamera_include_dirs():
   # the "usr/local" prefix is for recent Ubunut versions, which
   # install addon modules no longer along with the python core
   return [os.path.join(get_python_inc(), "gamera"),
           os.path.join(get_python_inc(prefix="/usr/local"), "gamera"),
           os.path.join(get_python_inc(), "../gamera"),
           "/usr/include/gamera"]



#
#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom, Karl MacMillan
#               2013      Fabian Schmitt
#               2010-2013 Christoph Dalitz
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#  
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""The transformation module contains plugins for geometric transformations
like rotating or resizing."""

from gamera.plugin import * 
try:
  from gamera.core import RGBPixel
except Exception:
  def RGBPixel(*args):
    pass
from gamera.gui import has_gui
from gamera.util import warn_deprecated
from gamera.args import NoneDefault
import sys
import _transformation

class rotate(PluginFunction):
    """
    Rotates an image.

    *angle*
      The angle of rotation (in degrees)

    *bgcolor*
      The color to use for pixels outside of the original image bounds.
      When *bgcolor* is ``None``, white is used.

    *order*
      The order of the spline used for interpolation.  Must be between 1 - 3.
    """
    category = "Transformation"
    self_type = ImageType(ALL)    
    return_type = ImageType(ALL)
    args = Args([Float("angle"), Pixel("bgcolor", default=NoneDefault), Int("order", range=(1,3), default=1)])
    args.list[0].rng = (-180,180)
    doc_examples = [(RGB, 32.0, RGBPixel(255, 255, 255), 3), (COMPLEX, 15.0, 0.0j, 3)]
    author = u"Michael Droettboom (With code from VIGRA by Ullrich K\u00f6the)"
    def __call__(self, angle, bgcolor=None, order=1):
      if (bgcolor == None):
          bgcolor = self.white()
      return _transformation.rotate(self, angle, bgcolor, order)
    __call__ = staticmethod(__call__)
    
class resize(PluginFunction):
    """
    Returns a resized copy of an image. In addition to size, the type
    of interpolation can be specified, with a tradeoff between speed
    and quality.

    If you need to maintain the aspect ratio of the original image,
    consider using scale_ instead.

    *dim*
      The size of the resulting image.

    *interp_type* [None|Linear|Spline]
      The type of interpolation used to resize the image.  Each option
      is progressively higher quality, yet slower.

    .. _scale: #scale
    """
    category = "Transformation"
    self_type = ImageType(ALL)
    args = Args([Dim("dim"), Choice("interp_type", ["None", "Linear", "Spline"])])
    return_type = ImageType(ALL)

class scale(PluginFunction):
    """
    Returns a scaled copy of the image. In addition to scale, the type
    of interpolation can be specified, with a tradeoff between speed
    and quality.

    If you need to change the aspect ratio of the original image,
    consider using resize_ instead.

    *scale*
      A scaling factor.  Values greater than 1 will result in a larger image.
      Values less than 1 will result in a smaller image.

    *interp_type* [None|Linear|Spline]
      The type of interpolation used to resize the image.  Each option is
      progressively higher quality, yet slower.

    .. _resize: #resize
    """
    category = "Transformation"
    self_type = ImageType(ALL)
    args= Args([Real("scaling"),
                Choice("interp_type", ["None", "Linear", "Spline"])])
    return_type = ImageType(ALL)
    doc_examples = [(RGB, 0.5, 2), (RGB, 2.0, 2)]

class shear_row(PluginFunction):
    """
    Shears a given row by a given amount.

    *row*
      The row number to shear.

    *distance*
      The number of pixels to move the row.  Positive values
      move the row to the right.  Negative values move the row
      to the left.
    """
    category = "Transformation"
    self_type = ImageType(ALL)
    args = Args([Int('row'), Int('distance')])
    doc_examples = [(ONEBIT, 50, 10)]

class shear_column(PluginFunction):
    """
    Shears a given column by a given amount.

    *column*
      The column number to shear.

    *distance*
      The number of pixels to move the column.  Positive values
      move the column downward.  Negative values move the column
      upward.
    """
    category = "Transformation"
    self_type = ImageType(ALL)
    args = Args([Int('column'), Int('distance')])
    doc_examples = [(ONEBIT, 50, 10)]

class mirror_horizontal(PluginFunction):
    """
    Flips the image across the horizontal (*x*) axis.
    """
    category = "Transformation"
    self_type = ImageType(ALL)
    doc_examples = [(RGB,)]

class mirror_vertical(PluginFunction):
    """
    Flips the image across the vertical (*y*) axis.
    """
    category = "Transformation"
    self_type = ImageType(ALL)
    doc_examples = [(RGB,)]

class grey_convert(PluginFunction):
    """
    Converts grey values to the new values provided in the vector *newgrey*,
    i.e. *val* is replaced by *newgrey[val]*.

    As this is only implemented for greyscale images, the vector *newgrey*
    must be of size 256 and all entries must be in [0,255].
    """
    category = "Transformation"
    author = "Fabian Schmitt and Christoph Dalitz"
    self_type = ImageType([GREYSCALE])
    return_type = ImageType([GREYSCALE])
    args = Args([IntVector ("newgrey")])
    doc_examples = [(GREYSCALE,range(255,-1,-1))]

class TransformationModule(PluginModule):
    cpp_headers=["transformation.hpp"]
    category = "Transformation"
    functions = [rotate, resize, scale,
                 shear_row, shear_column,
                 mirror_horizontal, mirror_vertical, grey_convert]
    author = "Michael Droettboom, Karl MacMillan, and Christoph Dalitz"
    url = "http://gamera.sourceforge.net/"
module = TransformationModule()




# -*- mode: python; indent-tabs-mode: nil; tab-width: 3 -*-
# vim: set tabstop=3 shiftwidth=3 expandtab:
#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom, Karl MacMillan
#               2010      Christoph Dalitz
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from pyplate import *
from os import path
import os
import sys
import re
from distutils.core import Extension
from distutils.dep_util import newer
from distutils import sysconfig
from gamera import paths, args_wrappers

global std_import
global plugins_to_ignore

# magic_import and magic_import_setup
#
# This allows us to ignore a list of modules passed into
# magic_import_setup. generate_plugin uses this to prevent
# the loading of C++ modules that may not exist yet during
# the build process.
def magic_import(name, globals_={}, locals_={}, fromlist=[], level=-1):
   if fromlist != None and "core" in fromlist:
      fromlist = list(fromlist)
      fromlist.remove("core")

   for x in plugins_to_ignore:
      if name == x:
         return None

   if float(sys.version[0:3]) < 2.45:
      return std_import(name, globals_, locals_, fromlist)
   else:
      return std_import(name, globals_, locals_, fromlist, level)

def magic_import_setup(ignore):
   global plugins_to_ignore
   global std_import
   plugins_to_ignore = ignore
   # Save the standard __import__ function so we can chain to it
   std_import = __builtins__['__import__']
   # Override the __import__ function with our new one
   __builtins__['__import__'] = magic_import

def restore_import():
   global std_import
   __builtins__['__import__'] = std_import

template = Template("""
  [[exec from os import path]]
  [[exec from enums import *]]
  [[exec from plugin import *]]
  [[exec from util import get_pixel_type_name]]

  [[# This should be included first in order to avoid libpng.h/setjmp.h problems. #]]
  [[if module.__class__.__name__ == "PngSupportModule"]]
    #include <png.h>
    #undef _POSIX_C_SOURCE
    #undef _XOPEN_SOURCE
  [[end]]

  #include \"gameramodule.hpp\"
  #include \"knnmodule.hpp\"

  [[# include the headers that the module needs #]]
  [[for header in module.cpp_headers]]
    #include \"[[header]]\"
  [[end]]

  [[# Standard headers used in the plugins #]]
  #include <string>
  #include <stdexcept>
  #include \"Python.h\"
  #include <list>

  using namespace Gamera;
  [[for x in module.cpp_namespaces]]
    using namespace [[x]];
  [[end]]

  [[# Generate the plugin path and module name from the filename. #]]
  [[# The module name for our purposes will be prefixed with an underscore #]]
  [[exec plug_path, filename = path.split(__file__)]]
  [[exec module_name = '_' + filename.split('.')[0] ]]

  [[# Declare all of the functions - because this is a C++ file we have to #]]
  [[# declare the functions as C functions so that Python can access them #]]
  extern \"C\" {
#ifndef _MSC_VER
    void init[[module_name]](void);
#endif
    [[for function in module.functions]]
      [[if not function.pure_python]]
        static PyObject* call_[[function.__name__]](PyObject* self, PyObject* args);
      [[end]]
    [[end]]
  }

  [[# Create the list of methods for the module - the name of the function #]]
  [[# is derived from the name of the class implementing the function - #]]
  [[# also, the function name is prepended with call_ so that there are no clashes #]]
  [[# with the real plugin functions #]]
  static PyMethodDef [[module_name]]_methods[] = {
    [[for function in module.functions]]
      [[if not function.pure_python]]
        { CHAR_PTR_CAST \"[[function.__name__]]\",
          call_[[function.__name__]], METH_VARARGS,
          CHAR_PTR_CAST [[function.escape_docstring()]]
        },
      [[end]]
    [[end]]
    { NULL }
  };

  [[# Each module can declare several functions so we loop through and generate wrapping #]]
  [[# code for each function #]]
  [[for function in module.functions]]
    [[if not function.pure_python]]
      static PyObject* call_[[function.__name__]](PyObject* self, PyObject* args) {
      [[# this holds the self argument - note that the self passed into the function will #]]
      [[# be Null because this functions is not actually bound to an object #]]

      PyErr_Clear();
      [[if function.self_type == None]]
        [[exec args = function.args.list]]
      [[else]]
        [[exec args = [function.self_type] + function.args.list]]
        [[exec function.self_type.name = 'self']]
      [[end]]
      [[# for each argument insert the appropriate conversion code into the string that will #]]
      [[# be passed to PyArg_ParseTuple and create a variable to hold the result. #]]
      [[if function.return_type != None]]
        [[exec function.return_type.name = 'return']]
        [[exec function.return_type.convert_from_PyObject = True]]
        [[if not function.feature_function]]
          [[function.return_type.declare()]]
        [[end]]
      [[end]]
      [[exec pyarg_format = '']]
      [[for arg in args]]
        [[exec pyarg_format += arg.arg_format]]
        [[arg.declare()]]
      [[end]]

      [[# Now that we have all of the arguments and variables for them we can parse #]]
      [[# the argument tuple. #]]
      [[if function.feature_function]]
         int offset = -1;
         if (PyArg_ParseTuple(args, CHAR_PTR_CAST \"O|i:[[function.__name__]]\",&[[function.self_type.pysymbol]], &offset) <= 0)
           return 0;
      [[else]]
         [[if pyarg_format != '']]
           if (PyArg_ParseTuple(args, CHAR_PTR_CAST \"[[pyarg_format]]:[[function.__name__]]\"
           [[for arg in args]]
             ,
             &[[arg.pysymbol]]
           [[end]]
           ) <= 0)
           return 0;
         [[end]]
      [[end]]

      [[for arg in args]]
        [[arg.from_python()]]
      [[end]]

      [[if function.feature_function]]
         feature_t* feature_buffer = 0;
         if (offset < 0) {
           feature_buffer = new feature_t[ [[function.return_type.length]] ];
         } else {
           if (self_arg->features_len < offset + [[function.return_type.length]]) {
             PyErr_Format(PyExc_ValueError, \"Offset as given (%d) will cause data to be written outside of array of length (%d).  Perhaps the feature array is not initialised?\", offset, (int)self_arg->features_len);
             return 0;
           }
           feature_buffer = self_arg->features + offset;
         }
         [[args[0].call(function, args[1:], [])]]
      [[else]]
        try {
          [[if len(args)]]
            [[args[0].call(function, args[1:], [])]]
          [[else]]
            [[if function.return_type != None]]
              [[function.return_type.symbol]] =
            [[end]]
            [[function.__name__]]([[if function.progress_bar]]ProgressBar("[[function.progress_bar]]")[[else]][[end]]);
          [[end]]
        } catch (std::exception& e) {
          PyErr_SetString(PyExc_RuntimeError, e.what());
          return 0;
        }
      [[end]]

      [[if function.feature_function]]
         if (offset < 0) {
           PyObject* str = PyString_FromStringAndSize((char*)feature_buffer, [[function.return_type.length]] * sizeof(feature_t));
           if (str != 0) {
              [[# This is pretty expensive, but simple #]]
              PyObject* array_init = get_ArrayInit();
              if (array_init == 0)
                return 0;
              PyObject* array = PyObject_CallFunction(
                    array_init, (char *)\"sO\", (char *)\"d\", str);
              Py_DECREF(str);
              delete[] feature_buffer;
              return array;
           } else {
             delete[] feature_buffer;
             return 0;
           }
         } else {
           Py_INCREF(Py_None);
           return Py_None;
         }
      [[else]]
        [[for arg in function.args]]
          [[arg.delete()]]
        [[end]]
        [[if function.return_type == None]]
          Py_INCREF(Py_None);
          return Py_None;
        [[else]]
          [[if isinstance(function.return_type, (ImageType, Class))]]
            if ([[function.return_type.symbol]] == NULL) {
              if (PyErr_Occurred() == NULL) {
                Py_INCREF(Py_None);
                return Py_None;
               } else
                return NULL;
            } else {
              [[function.return_type.to_python()]]
              return return_pyarg;
            }
          [[else]]
            [[function.return_type.to_python()]]
            return return_pyarg;
          [[end]]
        [[end]]
      [[end]]
      }
    [[end]]
  [[end]]

  DL_EXPORT(void) init[[module_name]](void) {
    Py_InitModule(CHAR_PTR_CAST \"[[module_name]]\", [[module_name]]_methods);
  }
  """)

def generate_plugin(plugin_filename, location, compiling_gamera,
                    extra_compile_args=[], extra_link_args=[], libraries=[],
                    define_macros=[]):
  from gamera import gamera_setup

  plug_path, filename = path.split(plugin_filename)
  module_name = filename.split('.')[0]
  cpp_filename = path.join(plug_path, "_" + module_name + ".cpp")

  regenerate = False
  if newer(plugin_filename, cpp_filename) or '-f' in sys.argv:
    regenerate = True

  sys.path.append(plug_path)

  #import plugin
  plugin_module = __import__(module_name)
  if not hasattr(plugin_module, 'module'):
     return None
  if plugin_module.module.pure_python:
     return None

  # see if any of the header files have changed since last time
  # we compiled
  include_dirs = (["include", plug_path, "include/plugins"] +
                  plugin_module.module.cpp_include_dirs)
  if not compiling_gamera:
     include_dirs.extend(gamera_setup.get_gamera_include_dirs())
  if not regenerate:
    for header in plugin_module.module.cpp_headers:
      found_header = 0
      for include_dir in include_dirs:
        header_filename = path.join(include_dir, header)
        if path.exists(header_filename):
          found_header = 1
          if newer(header_filename, cpp_filename):
            regenerate = True
            break
          break
      if regenerate:
        break

  if regenerate:
    print "generating wrappers for", module_name, "plugin"
    template.execute_file(cpp_filename, plugin_module.__dict__)
  else:
    print "skipping wrapper generation for", module_name, "plugin (output up-to-date)"

  # make the a distutils extension class for this plugin
  cpp_files = [cpp_filename]
  for file in plugin_module.module.cpp_sources:
    cpp_files.append(file)

  extra_libraries = plugin_module.module.extra_libraries
  # This is to make up for a bug in distutils.
  if '--compiler=mingw32' in sys.argv or not sys.platform == 'win32':
     if "stdc++" not in extra_libraries:
        extra_libraries.append("stdc++")
  return Extension(location + "._" + module_name, cpp_files,
                   include_dirs=include_dirs,
                   library_dirs=plugin_module.module.library_dirs,
                   libraries=extra_libraries,
                   extra_compile_args=plugin_module.module.extra_compile_args + extra_compile_args,
                   extra_link_args=plugin_module.module.extra_link_args + extra_link_args,
                   define_macros=plugin_module.module.define_macros + define_macros,
                   extra_objects=plugin_module.module.extra_objects)




from gamera.core import *
init_gamera()

def test_save_image():
   def _test_save_image(name, storage = DENSE):
      for ext in ['png', 'tiff']:
         image = load_image("data/%s_generic.tiff" % name, storage)
         image.save_image("tmp/%s_test.%s" % (name, ext))
         image2 = load_image("tmp/%s_test.%s" % (name, ext))
         assert image._to_raw_string() == image2._to_raw_string()

   # TODO: Add Grey16
   for type in ["OneBit", "GreyScale", "RGB"]:
      _test_save_image(type)
   _test_save_image("OneBit", RLE)




# -*- mode: python; indent-tabs-mode: nil; tab-width: 3 -*-
# vim: set tabstop=3 shiftwidth=3 expandtab:
#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom,
#                          and Karl MacMillan
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import gzip, os, os.path, cStringIO
import warnings
from weakref import proxy
from xml.parsers import expat

import core, util
from util import word_wrap, ProgressFactory, is_image_list
from gamera.plugins import runlength
from gamera.symbol_table import SymbolTable
from config import config

config.add_option(
   "", "--xml-encoding", action="store", default="utf-8",
   help='[xml] Character encoding to use when saving XML files')

GAMERA_XML_FORMAT_VERSION = 2.0

_classification_state_to_name = {
   core.UNCLASSIFIED: "UNCLASSIFIED",
   core.MANUAL:       "MANUAL",
   core.HEURISTIC:    "HEURISTIC",
   core.AUTOMATIC:    "AUTOMATIC" }
def classification_state_to_name(state):
   return _classification_state_to_name[state]

_classification_state_to_number = {
   "UNCLASSIFIED":    core.UNCLASSIFIED,
   "MANUAL":          core.MANUAL,
   "HEURISTIC":       core.HEURISTIC,
   "AUTOMATIC":       core.AUTOMATIC }
def classification_state_to_number(state):
   return _classification_state_to_number[state]

_saveable_types = {
   'int':   int,
   'str':   str,
   'float': float }

extensions = "XML files (*.xml;*.xml.gz)|*.xml;*.xml.gz|All files|*"

class XMLError(Exception):
   pass

################################################################################
# SAVING
################################################################################

class WriteXML:
   def __init__(self, glyphs=[], symbol_table=[], with_features=True):
      self.glyphs = glyphs
      if (not (isinstance(symbol_table, SymbolTable) or
               util.is_string_or_unicode_list(symbol_table))):
         raise XMLError(
            "symbol_table argument to WriteXML must be of type SymbolTable or a list of strings.")
      self.symbol_table = symbol_table
      self.with_features = with_features

   def write_filename(self, filename, with_features=None):
      if not with_features is None:
         self.with_features = with_features
      if not os.path.exists(os.path.split(os.path.abspath(filename))[0]):
         raise XMLError(
            "Cannot create a file at '%s'." %
            os.path.split(os.path.abspath(filename))[0])
      if filename.endswith('gz'):
         fd = gzip.open(filename, 'w')
      else:
         fd = open(filename, 'w')
      self.write_stream(fd)

   def string(self):
      stream = cStringIO.StringIO()
      self.write_stream(stream)
      return stream.getvalue()

   def write_stream(self, stream=None):
      if stream == None:
         return self.string()
      self._write_core(stream)

   def _write_core(self, stream, indent=0):
      progress = util.ProgressFactory("Saving XML...", 1, numsteps=32)
      try:
         self._write_symbol_table(stream, self.symbol_table, indent=indent)
         if isinstance(self.glyphs, core.ImageBase):
            progress.add_length(1)
            self._write_glyph(stream, self.glyphs, indent=indent)
            progress.step()
         else:
            progress.add_length(len(self.glyphs))
            self._write_glyphs(stream, self.glyphs, progress, indent=indent)
      finally:
         progress.kill()

   def _write_symbol_table(self, stream, symbol_table, indent=0):
      encoding = config.get("xml_encoding")
      if (not isinstance(symbol_table, SymbolTable) and
          util.is_string_or_unicode_list(symbol_table)):
         symbols = symbol_table
      else:
         symbols = symbol_table.symbols.keys()
      if len(symbols):
         symbols.sort()
         word_wrap(stream, '<symbols>', indent)
         indent += 1
         for x in symbols:
            word_wrap(stream, '<symbol name="%s"/>' % x.encode(encoding), indent)
         indent -= 1
         word_wrap(stream, '</symbols>', indent)

   def _write_glyphs(self, stream, glyphs, progress, indent=0):
      if len(glyphs):
         word_wrap(stream, '<glyphs>', indent)
         indent += 1
         for i, glyph in util.enumerate(glyphs):
            self._write_glyph(stream, glyph, indent)
            progress.step()
         indent -= 1
         word_wrap(stream, '</glyphs>', indent)

   def _write_glyph(self, stream,  glyph, indent=0):
      tag = ('<glyph uly="%s" ulx="%s" nrows="%s" ncols="%s">' %
             (glyph.ul_y, glyph.ul_x, glyph.nrows, glyph.ncols))
      word_wrap(stream, tag, indent)
      indent += 1
      word_wrap(
         stream,
         '<ids state="%s">' %
         classification_state_to_name(glyph.classification_state),
         indent)
      indent += 1
      for confidence, id in glyph.id_name:
         word_wrap(stream, '<id name="%s" confidence="%f"/>' %
                   (id, confidence), indent)
      indent -= 1
      word_wrap(stream, '</ids>', indent)
      word_wrap(stream, '<data>', indent)
      word_wrap(stream, glyph.to_rle(), indent+1)
      word_wrap(stream, '</data>', indent)
      feature_functions = glyph.feature_functions[0]
      if self.with_features and len(feature_functions):
         word_wrap(stream,
                   '<features scaling="%s">' % str(glyph.scaling),
                   indent)
         indent += 1
         feature_no = 0
         for name, function in feature_functions:
            word_wrap(stream,
                      '<feature name="%s">' % name,
                      indent)
            length = function.return_type.length
            word_wrap(stream,
                      [x for x in
                       glyph.features[feature_no:feature_no+length]],
                      indent + 1)
            feature_no += length
            word_wrap(stream,
                      '</feature>',
                      indent)
         indent -= 1
         word_wrap(stream, '</features>', indent)
      properties = glyph.properties.items()
      properties.sort()
      for key, val in properties:
         if not val is None:
            word_wrap(stream, '<property name="%s" type="%s">%s</property>' %
                      (key, type(val).__name__, str(val)), indent)
      indent -= 1
      word_wrap(stream, '</glyph>', indent)

class WriteXMLFile(WriteXML):
   def write_stream(self, stream=None):
      if stream == None:
         return self.string()
      self.stream = stream
      encoding = config.get("xml_encoding")
      self.stream.write('<?xml version="1.0" encoding="%s"?>\n' % encoding)
      self.stream.write('<gamera-database version="%s">\n' %
                        str(GAMERA_XML_FORMAT_VERSION))
      self._write_core(stream, indent=1)
      self.stream.write('</gamera-database>\n')


################################################################################
# LOADING
################################################################################

class LoadXML:
   def __init__(self, parts = ['symbol_table', 'glyphs']):
      self._start_elements = {}
      self._end_elements = {}
      self._stream_length = 0
      self._parts = parts
      self._progress_value = 0

   def try_type_convert(self, dictionary, key, typename, tagname):
      try:
         return typename(dictionary[key])
      except KeyError:
         raise XMLError(
            "XML ValueError: <%s> tag does not have a required attribute '%s'." %
            (tagname, key))
      except TypeError:
         raise XMLError(
            'XML ValueError: <%s %s="%s" ... is not of the correct type' %
            (tagname, key, dictionary[key]))
      
   def parse_filename(self, filename):
      try:
         self._stream_length = os.stat(filename).st_size
      except OSError, e:
         raise XMLError(str(e))
      if filename.endswith('gz'):
         fd = gzip.open(filename, 'r')
      else:
         fd = open(filename, 'r')
      try:
         return self.parse_stream(fd)
      except Exception, e:
         raise XMLError(str(e))

   def parse_string(self, s):
      self._stream_length = len(s)
      stream = cStringIO.StringIO(s)
      return self.parse_stream(stream)

   def parse_stream(self, stream):
      self._setup_handlers()
      self._parser = expat.ParserCreate()
      self._parser.StartElementHandler = self._start_element_handler
      self._parser.EndElementHandler = self._end_element_handler
      self._stream = stream
      self._progress = util.ProgressFactory("Loading XML...", self._stream_length, numsteps=32)
      try:
         try:
            self._parser.ParseFile(stream)
         except expat.ExpatError, e:
            raise
      finally:
         self._progress.kill()
         self._remove_handlers()
         self._parser.StartElementHandler = None
         self._parser.EndElementHandler = None
         del self._parser
      return self
   
   def add_start_element_handler(self, name, func):
      self._start_elements[name] = func

   def remove_start_element_handler(self, name):
      del self._start_elements[name]

   def _start_element_handler(self, name, attributes):
      try:
         self._start_elements[name](attributes)
      except KeyError:
         pass

   def add_end_element_handler(self, name, func):
      self._end_elements[name] = func

   def remove_end_element_handler(self, name):
      del self._end_elements[name]

   def _end_element_handler(self, name):
      try:
         self._end_elements[name]()
      except KeyError:
         pass

   def _update_progress(self):
      if self._stream_length:
         self._progress.update(self._stream.tell(), self._stream_length)
      else:
         self._progress_value += 1
         self._progress_value %= 50
         self._progress.update(self._progress_value, 50)

   def _append_glyph_to_glyphs(self, glyph):
      self.glyphs.append(glyph)

   def _setup_handlers(self):
      self.symbol_table = SymbolTable()
      self.glyphs = []
      self.add_start_element_handler('gamera-database', self._tag_start_gamera_database)
      self.add_end_element_handler('gamera-database', self._tag_end_gamera_database)

   def _remove_handlers(self):
      self.remove_start_element_handler('gamera-database')
      self.remove_end_element_handler('gamera-database')

   def _tag_start_gamera_database(self, a):
      try:
         version = self.try_type_convert(a, 'version', float, 'gamera-database')
      except Exception:
         version = 1.0
      if version < GAMERA_XML_FORMAT_VERSION:
         raise XMLError(
            "The XML file is an old version, which can not be read " +
            "by this version of Gamera.")
      if 'symbol_table' in self._parts:
         self.add_start_element_handler('symbols', self._tag_start_symbols)
         self.add_end_element_handler('symbols', self._tag_end_symbols)
      if 'glyphs' in self._parts:
         self.add_start_element_handler('glyphs', self._tag_start_glyphs)
         self.add_end_element_handler('glyphs', self._tag_end_glyphs)

   def _tag_end_gamera_database(self):
      if 'symbol_table' in self._parts:
         self.remove_start_element_handler('symbols')
         self.remove_end_element_handler('symbols')
      if 'glyphs' in self._parts:
         self.remove_start_element_handler('glyphs')
         self.remove_end_element_handler('glyphs')

   def _tag_start_symbols(self, a):
      self.add_start_element_handler('symbol', self._tag_start_symbol)

   def _tag_end_symbols(self):
      self.remove_start_element_handler('symbol')
   
   def _tag_start_symbol(self, a):
      self.symbol_table.add(str(a['name']))
      self._update_progress()

   def _tag_start_glyphs(self, a):
      self._append_glyph = self._append_glyph_to_glyphs
      self.add_start_element_handler('glyph', self._tag_start_glyph)
      self.add_end_element_handler('glyph', self._tag_end_glyph)
      self.add_start_element_handler('features', self._tag_start_features)
      self.add_start_element_handler('ids', self._tag_start_ids)
      self.add_start_element_handler('id', self._tag_start_id)
      self.add_start_element_handler('data', self._tag_start_data)
      self.add_end_element_handler('data', self._tag_end_data)
      self.add_start_element_handler('property', self._tag_start_property)
      self.add_end_element_handler('property', self._tag_end_property)

   def _tag_end_glyphs(self):
      self._append_glyph = None
      for element in 'glyph features ids id data property'.split():
         self.remove_start_element_handler(element)
      for element in 'glyph data property'.split():
         self.remove_end_element_handler(element)

   def _tag_start_glyph(self, a):
      self._ul_y = self.try_type_convert(a, 'uly', int, 'glyph')
      self._ul_x = self.try_type_convert(a, 'ulx', int, 'glyph')
      self._nrows = self.try_type_convert(a, 'nrows', int, 'glyph')
      self._ncols = self.try_type_convert(a, 'ncols', int, 'glyph')
      self._scaling = 1.0
      self._id_name = []
      self._properties = {}
      self._data = None
      self._classification_state = core.UNCLASSIFIED

   def _tag_end_glyph(self):
      glyph = core.Image(core.Point(self._ul_x, self._ul_y),
                         core.Dim(self._ncols, self._nrows),
                         core.ONEBIT, core.DENSE)
      if not self._data is None:
         glyph.from_rle(str(u''.join(self._data)))
      glyph.classification_state = self._classification_state
      self._id_name.sort()
      glyph.id_name = self._id_name
      for key, val in self._properties.items():
         glyph.properties[key] = val
      glyph.scaling = self._scaling
      self._append_glyph(glyph)
      if not len(self.glyphs) & 0xf:
         self._update_progress()

   def _tag_start_ids(self, a):
      self._classification_state = self.try_type_convert(
         a, 'state', classification_state_to_number, 'ids')

   def _tag_start_id(self, a):
      confidence = self.try_type_convert(
         a, 'confidence', float, 'id')
      name = self.try_type_convert(
         a, 'name', unicode, 'id')
      self._id_name.append((confidence, name.encode()))

   def _tag_start_features(self, a):
      self._scaling = self.try_type_convert(
         a, 'scaling', float, 'features')

   def _tag_start_data(self, a):
      self._data = []
      self._parser.CharacterDataHandler = self.add_data

   def _tag_end_data(self):
      self._parser.CharacterDataHandler = None

   def add_data(self, data):
      self._data.append(data)

   def _tag_start_property(self, a):
      self._property_name = self.try_type_convert(
         a, 'name', str, 'property')
      self._property_type = self.try_type_convert(
         a, 'type', str, 'property')
      self._property_value = []
      self._parser.CharacterDataHandler = self.add_property_value

   def _tag_end_property(self):
      data = u''.join(self._property_value)
      if _saveable_types.has_key(self._property_type):
         self._properties[self._property_name] = \
            _saveable_types[self._property_type](data)
      else:
         self._properties[self._property_name] = data
      self._parser.CharacterDataHandler = None

   def add_property_value(self, data):
      self._property_value.append(data)

def glyphs_from_xml(filename, feature_functions = None):
   """**glyphs_from_xml** (*filename*, *feature_functions* = ``None``)

Return a list of glyphs from a Gamera XML file."""
   glyphs = LoadXML().parse_filename(filename).glyphs
   if not feature_functions is None:
      from gamera.plugins import features
      features.generate_features_list(glyphs, feature_functions)
   return glyphs

def glyphs_with_features_from_xml(filename, feature_functions = None):
   """**glyphs_with_features_from_xml** (*filename*, *feature_functions* = ``None``)

Loads glyphs from a Gamera XML file, and then generates features
for all of those glyphs.  The set of features can be specified with the
*feature_functions* argument (which defaults to all features)."""
   warnings.warn("Use glyphs_from_xml with a feature descriptor instead of glyphs_with_features_from_xml.", DeprecationWarning)
   if feature_functions == None:
      feature_functions = 'all'
   return glyphs_from_xml(filename, feature_functions)

def glyphs_to_xml(filename, glyphs, with_features=True):
   """**glyphs_to_xml** (*filename*, *glyphs*, *with_features* = ``True``)

Saves the given list of glyphs to a Gamera XML file.

*with_features*
  When set to ``True``, features generated on the image are saved to the XML file.
"""
   WriteXMLFile(glyphs, with_features=with_features).write_filename(filename) 	 
 
class StripTag:
   # This is a ridiculous implementation that probably deserves some
   # explanation.  I want to use a full-fledged XML parser, (even though this
   # could probably be done with Regex's or something), for robustness.
   # However, when parsing a file with Expat, there's no way of getting the
   # original XML back, and .GetInputContext() seems to be broken in Py2.3.3
   # AFAICT.  Therefore, what this does is two passes:
   #   1) go through storing the indices of start and end tags
   #   2) use that information to generate an excerpted file
   def __init__(self, input_filename, output_filename, tag_name):
      if input_filename == output_filename:
         raise ValueError("Input and output filenames must be different.")
      self._input_filename = input_filename
      self._output_filename = output_filename
      self._tag_name = tag_name
      self.parse()
      self.excerpt()

   def open_input_file(self):
      try:
         if self._input_filename.endswith('gz'):
            self._input = gzip.open(self._input_filename, 'r')
         else:
            self._input = open(self._input_filename, 'r')
      except Exception, e:
         raise XMLError("Couldn't open input file '%s': %s" %
                        (self._input_filename, str(e)))

   def open_output_file(self):
      try:
         if self._output_filename.endswith('gz'):
            self._output = gzip.open(self._output_filename, 'w')
         else:
            self._output = open(self._output_filename, 'w')
      except Exception, e:
         raise XMLError("Couldn't open output file '%s': %s" %
                        (self._output_filename, str(e)))

   def parse(self):
      self.open_input_file()
      self._parser = expat.ParserCreate()
      self._parser.StartElementHandler = self.start_element_handler
      self._parser.EndElementHandler = self.end_element_handler
      self._parser.DefaultHandler = self.default_handler
      self._last_was_end = False
      self._indices = []
      try:
         try:
            self._parser.ParseFile(self._input)
         except expat.ExpatError, e:
            raise
      finally:
         self._parser.StartElementHandler = None
         self._parser.EndElementHandler = None
         del self._parser
      self._input.close()

   def start_element_handler(self, name, attributes):
      if self._last_was_end:
         self._indices.append(self._parser.ErrorByteIndex)
      self._last_was_end = False
      if name == self._tag_name:
         self._indices.append(self._parser.ErrorByteIndex)

   def end_element_handler(self, name):
      if self._last_was_end:
         self._indices.append(self._parser.ErrorByteIndex)
      if name == self._tag_name:
         self._last_was_end = True
      else:
         self._last_was_end = False

   def default_handler(self, data):
      if self._last_was_end:
         self._indices.append(self._parser.ErrorByteIndex)
      self._last_was_end = False

   def excerpt(self):
      self.open_input_file()
      self.open_output_file()
      pos = 0
      for i in range(0, len(self._indices), 2):
         start, end = self._indices[i], self._indices[i+1]
         self._output.write(self._input.read(start - pos))
         pos = end
         self._input.seek(pos)
      self._output.write(self._input.read())
      self._input.close()
      self._output.close()

def strip_features(input_filename, output_filename):
   """**strip_features** (*input_filename*, *output_filename*)

Strips the features from a Gamera XML file.  Provided mainly to reduce filesizes for
files created before saving features was an option.

*input_filename*
  The input Gamera XML filename

*output_filename*
  The output Gamera XML filename"""
   StripTag(input_filename, output_filename, 'features')
   
   



#!/usr/bin/env python
# -*- mode: python; indent-tabs-mode: nil; tab-width: 3 -*-
# vim: set tabstop=3 shiftwidth=3 expandtab:
#
# Copyright (C) 2003-2008 Michael Droettboom
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
"""Installs an icon on the user's Desktop when Gamera is installed."""

import sys, os.path, os

def install_desktop_icon():
  try:
    try:
      import wx
    except Exception:
      import wxPython.wx
  except ImportError:
    print "wxPython was not found.  Please make sure that wxPython is installed before running Gamera."

  try:
    #desktop_path = get_special_folder_path("CSIDL_COMMON_DESKTOPDIRECTORY")
    desktop_path = get_special_folder_path("CSIDL_DESKTOPDIRECTORY")
  except OSError:
    print "Can't find the desktop!"
    return
  if os.path.exists(desktop_path):
    print "Creating desktop icon..."
    desktop_icon = os.path.join(desktop_path, "Gamera.py")
    fd = open(desktop_icon, 'w')
    fd.write("""
print "Loading GAMERA..."
import sys
try:
   import wxversion
except ImportError:
   print "Please ensure that wxPython is installed before running Gamera."
   print "Press <ENTER> to exit."
   x = raw_input()
   sys.exit()
try:
   from gamera.gui import gui
   gui.run()
except Exception:
   import traceback
   print "Gamera made a fatal error:"
   print
   traceback.print_exc()
   print
   print "Press <ENTER> to exit."
   x = raw_input()
   sys.exit(1)""")
    fd.close()
    file_created(desktop_icon)
    print "done."

if sys.argv[1] == '-install':
  install_desktop_icon()
elif sys.argv[1] == '-remove':
  pass
else:
  print "This script should only be run by the Windows installer."



#
# Copyright (C) 2012 Tobias Bolten
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import wx

try:
    from agw import floatspin as FS
except ImportError:
    try:
        import wx.lib.agw.floatspin as FS
    except ImportError:
        raise RuntimeError("Biollante requires at least wxPython 2.8.11")

from gamera.gui.gaoptimizer.ExpertSettingPanel import *

#-------------------------------------------------------------------------------
class SelectionPanel(ExpertSettingPanel):
#-------------------------------------------------------------------------------
    #---------------------------------------------------------------------------
    def __init__(self, parent, id):
    #---------------------------------------------------------------------------
        ExpertSettingPanel.__init__(self, parent, id)

        sizer = wx.GridBagSizer(hgap=5, vgap=5)
        self.SetSizer(sizer)
        
        # roulette wheel selection
        self.roulettWheel = wx.RadioButton(self, -1, "Roulette wheel", \
            name="rouletteWheel")
        sizer.Add(self.roulettWheel, pos=(0,0), \
            flag = wx.LEFT | wx.RIGHT | wx.TOP | wx.EXPAND, border=10)

        self.genericWidgets.append(self.roulettWheel)

        # linear scaled fitness
        self.roulettWheelScaled = wx.RadioButton(self, -1, "Roulette wheel (lin. scaled)", \
            name="rouletteWheelScaled")
        self.roulettWheelScaled.SetValue(True)
        sizer.Add(self.roulettWheelScaled, pos=(1,0), \
            flag = wx.LEFT | wx.RIGHT | wx.EXPAND, border=10)

        self.roulettWheelPreasure = FS.FloatSpin(self, -1, min_val=0.0, max_val=2.0, \
            increment=0.01, value=2.0)
        self.roulettWheelPreasure.SetFormat("%f")
        self.roulettWheelPreasure.SetDigits(2)
        self.roulettWheelPreasure.SetToolTipString("The selective pressure")
        sizer.Add(self.roulettWheelPreasure, pos=(1,1), \
            flag = wx.LEFT | wx.RIGHT | wx.EXPAND, border=10)

        self.genericWidgets.append(self.roulettWheelScaled)
        self.AddChildToParent(self.roulettWheelScaled, self.roulettWheelPreasure)

        # stochastic universal sampling
        self.stochUniSampling = wx.RadioButton(self, -1, "Stochastic universal sampling", \
            name="stochUniSampling")
        sizer.Add(self.stochUniSampling, pos=(2,0), \
            flag = wx.LEFT | wx.RIGHT | wx.EXPAND, border=10)

        self.genericWidgets.append(self.stochUniSampling)

        # rank selection
        self.rankSelection = wx.RadioButton(self, -1, "Rank selection", name="rankSelection")
        sizer.Add(self.rankSelection, pos=(3,0), \
            flag = wx.LEFT | wx.RIGHT | wx.EXPAND, border=10)
            
        self.rankSelectionPreasure = FS.FloatSpin(self, -1, min_val=0.0, max_val=2.0, \
            increment=0.01, value=2.0)
        self.rankSelectionPreasure.SetFormat("%f")
        self.rankSelectionPreasure.SetDigits(2)
        self.rankSelectionPreasure.Disable()
        self.rankSelectionPreasure.SetToolTipString("The selective pressure")
        sizer.Add(self.rankSelectionPreasure, pos=(3,1), \
            flag = wx.LEFT | wx.RIGHT | wx.EXPAND, border=10)
            
        self.rankSelectionExponent = FS.FloatSpin(self, -1, min_val=0.0, max_val=2.0, \
            increment=0.01, value=1.0)
        self.rankSelectionExponent.SetFormat("%f")
        self.rankSelectionExponent.SetDigits(2)
        self.rankSelectionExponent.Disable()
        self.rankSelectionExponent.SetToolTipString("Exponent")
        sizer.Add(self.rankSelectionExponent, pos=(3,2), \
            flag = wx.LEFT | wx.RIGHT | wx.EXPAND, border=10)

        self.genericWidgets.append(self.rankSelection)
        self.AddChildToParent(self.rankSelection, self.rankSelectionPreasure)
        self.AddChildToParent(self.rankSelection, self.rankSelectionExponent)
            
        # tournament selection
        self.TournamentSelection = wx.RadioButton(self, -1, "Tournament selection", \
            name="tournamentSelection")
        sizer.Add(self.TournamentSelection, pos=(4,0), \
            flag = wx.LEFT | wx.RIGHT | wx.EXPAND, border=10)
            
        self.TournamentSelectionTsize = wx.SpinCtrl(self, -1, size=(100,-1), \
            min=2, max=25, value='3') # TODO: max = popSize
        self.TournamentSelectionTsize.Disable()
        self.TournamentSelectionTsize.SetToolTipString("Tournament size")
        sizer.Add(self.TournamentSelectionTsize, pos=(4,1), \
            flag = wx.LEFT | wx.RIGHT | wx.EXPAND, border=10)

        self.genericWidgets.append(self.TournamentSelection)
        self.AddChildToParent(self.TournamentSelection, self.TournamentSelectionTsize)

        # random selection
        self.randomSelect = wx.RadioButton(self, -1, "Random selection", \
            name="randomSelection")
        sizer.Add(self.randomSelect, pos=(5,0), \
            flag = wx.LEFT | wx.RIGHT | wx.BOTTOM | wx.EXPAND, border=10)

        self.genericWidgets.append(self.randomSelect)
            
        # bind the EVT_RADIOBUTTON to the RadioButtons
        self.BindEvent(wx.EVT_RADIOBUTTON, self.OnRadioButton, \
            [self.roulettWheel,
             self.roulettWheelScaled,
             self.stochUniSampling,
             self.rankSelection,
             self.TournamentSelection,
             self.randomSelect])



#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom,
#                         and Karl MacMillan
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from sys import stderr

try:
   import matplotlib
#    if (not hasattr(matplotlib, '__version__') or
#        matplotlib.__version__ not in ("0.73.1", "0.84", "0.90.1")):
#       print >>stderr, "WARNING: The version of matplotlib you have installed has not been officially"
#       print >>stderr, "tested with Gamera.  It may work fine, or you may experience strange"
#       print >>stderr, "problems using the matplotlib functionality.  Please include the"
#       print >>stderr, "version of your matplotlib (%s) in any bug reports to the Gamera" % (matplotlib.__version__)
#       print >>stderr, "developers.\n"
   try:
      matplotlib.use("WXAgg")
      from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg as FigureCanvas
      from matplotlib.backends import backend_wxagg as underlying_backend
   except ImportError:
      matplotlib.use("WX")
      from matplotlib.backends.backend_wx import FigureCanvasWx as FigureCanvas
      from matplotlib.backends import backend_wx as underlying_backend
   from matplotlib.backends import backend_wx
   from matplotlib import backend_bases
   from matplotlib.figure import Figure
   from matplotlib._pylab_helpers import Gcf
   import wx
   from gamera.gui import toolbar, gui_util, gamera_icons
except ImportError:
   def plot(*args, **kwargs):
      raise RuntimeError("Plotting is not supported because the optional matplotlib library\n"
                       "could not be found.\n\n"
                       "Download and install matplotlib from matplotlib.sourceforge.net,\n"
                       "then restart Gamera to have plotting support.")
   show_figure = plot
   matplotlib_installed = False
else:
   cursord = backend_wx.cursord
   
   class GameraPlotToolbar(backend_bases.NavigationToolbar2, toolbar.ToolBar):
      def __init__(self, parent, canvas):
         self.canvas = canvas
         self._idle = True
         self.statbar = None
         toolbar.ToolBar.__init__(self, parent)
         backend_bases.NavigationToolbar2.__init__(self, canvas)
         
      def _init_toolbar(self):
         load_bitmap = backend_wx._load_bitmap
         self.AddSimpleTool(10, gamera_icons.getIconHomeBitmap(),
                            'Reset original view', self.home)
         self.AddSimpleTool(20, gamera_icons.getIconBackBitmap(),
                            'Back navigational view', self.back)
         self.AddSimpleTool(30, gamera_icons.getIconForwardBitmap(),
                            'Forward navigational view', self.forward)
         self.AddSeparator()
         self.pan_button = self.AddSimpleTool(40, gamera_icons.getIconMoveBitmap(),
                                              'Pan/zoom mode', self.pan, True)
         self.zoom_button = self.AddSimpleTool(50, gamera_icons.getIconZoomViewBitmap(),
                                               'Zoom to rectangle', self.zoom, True)
         self.AddSeparator()
         self.AddSimpleTool(60, gamera_icons.getIconSaveBitmap(),
                            'Save plot contents to file', self.save)
         self.AddSimpleTool(70, gamera_icons.getIconPrinterBitmap(),
                            'Print', self.print_plot)
         
      def save(self, evt):
         filename = gui_util.save_file_dialog(self, self.canvas._get_imagesave_wildcards())
         if filename is not None:
            self.canvas.print_figure(filename)

      def print_plot(self, evt):
         printout = backend_wx.PrintoutWx(self.canvas)
         dialog_data = wx.PrintDialogData()
         if wx.VERSION < (2, 5):
            dialog_data.EnableHelp(False)
            dialog_data.EnablePageNumbers(False)
            dialog_data.EnableSelection(False)
         printer = wx.Printer(dialog_data)
         if not printer.Print(self, printout, True):
            if printer.GetLastError() == wx.PRINTER_ERROR:
               gui_util.message("A printing error occurred.")

      def zoom(self, evt):
         if evt.GetIsDown():
            self.pan_button.SetValue(False)
         else:
            self.zoom_button.SetValue(True)
         backend_bases.NavigationToolbar2.zoom(self, evt)
         
      def pan(self, evt):
         if evt.GetIsDown():
            self.zoom_button.SetValue(False)
         else:
            self.pan_button.SetValue(True)
         backend_bases.NavigationToolbar2.pan(self, evt)


      # This is all verbatim from backend_wx.py, which for various
      # multiple-ineheritance-related reasons can not just be directly
      # imported

      def set_cursor(self, cursor):
         cursor = wx.StockCursor(cursord[cursor])
         self.canvas.SetCursor( cursor )

      def release(self, event):
         try: del self.lastrect
         except AttributeError: pass

      def dynamic_update(self):
         d = self._idle
         self._idle = False
         if d:
            self.canvas.draw()
            self._idle = True

      def draw_rubberband(self, event, x0, y0, x1, y1):
         'adapted from http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/189744'
         canvas = self.canvas
         dc = wx.ClientDC(canvas)
         
         # Set logical function to XOR for rubberbanding
         dc.SetLogicalFunction(wx.XOR)
         
         # Set dc brush and pen
         # Here I set brush and pen to white and grey respectively
         # You can set it to your own choices
         
         # The brush setting is not really needed since we
         # dont do any filling of the dc. It is set just for 
         # the sake of completion.
         
         wbrush = wx.Brush(wx.Colour(255,255,255), wx.TRANSPARENT)
         wpen = wx.Pen(wx.Colour(200, 200, 200), 1, wx.SOLID)
         dc.SetBrush(wbrush)
         dc.SetPen(wpen)
         
         dc.ResetBoundingBox()
         dc.BeginDrawing()
         height = self.canvas.figure.bbox.height()
         y1 = height - y1
         y0 = height - y0
         
         if y1<y0: y0, y1 = y1, y0
         if x1<y0: x0, x1 = x1, x0
         
         w = x1 - x0
         h = y1 - y0
         
         rect = int(x0), int(y0), int(w), int(h)
         try: lastrect = self.lastrect
         except AttributeError: pass
         else: dc.DrawRectangle(*lastrect)  #erase last
         self.lastrect = rect
         dc.DrawRectangle(*rect)
         dc.EndDrawing()

      def set_status_bar(self, statbar):
         self.statbar = statbar

      def set_message(self, s):
         if self.statbar is not None: self.statbar.set_function(s)

   class GameraPlotDropTarget(wx.PyDropTarget):
      def __init__(self, figure):
         wx.PyDropTarget.__init__(self)
         self.df = wx.CustomDataFormat("Vector")
         self.data = wx.CustomDataObject(self.df)
         self.SetDataObject(self.data)
         self.figure = figure

      def OnEnter(self, *args):
         return wx.DragCopy
      
      def OnDrop(self, *args):
         return True
      
      def OnDragOver(self, *args):
         return wx.DragCopy
   
      def OnData(self, x, y, d):
         if self.GetData():
            data = eval(self.data.GetData())
            self.figure.axes[0].plot(data)
         return d

   class GameraPlotFrame(wx.Frame):
      def __init__(self, num, figure):
         self.num = num
         wx.Frame.__init__(self, None, -1, 'matplotlib Plot', size=(550, 350))
         self.figure = figure
         self.canvas = FigureCanvas(self, -1, self.figure)
         self.canvas.SetDropTarget(GameraPlotDropTarget(self.figure))
         statbar = backend_wx.StatusBarWx(self)
         self.SetStatusBar(statbar)
         self.toolbar = GameraPlotToolbar(self, self.canvas)
         self.toolbar.set_status_bar(statbar)

         box = wx.BoxSizer(wx.VERTICAL)
         box.Add(self.toolbar, 0, wx.EXPAND)
         box.Add(self.canvas, 1, wx.EXPAND)
         self.SetSizer(box)
         self.Fit()
         self.figmgr = GameraFigureManager(self.canvas, num, self)

      def GetToolBar(self):
         return self.toolbar

      def get_figure_manager(self):
         return self.figmgr

   _plot_num = 0
   def plot(*args):
      # Having inf values in the array raises a cryptic error
      # message from matplotlib
      inf = 1e300
      line = args[0]
      for x in line:
         if x > inf or x < -inf:
            raise ValueError("Line contains 'inf' or '-inf' values which can not be plotted.")

      figure = Figure()
      axis = figure.add_subplot(111)
      axis.plot(*args)
      show_figure(figure)
      return figure

   def show_figure(figure):
      display = GameraPlotFrame(0, figure)
      display.Show()
      return display

   # Everything below here is just to support pylab mode
   def show():
       for figwin in Gcf.get_all_fig_managers():
           figwin.frame.Show()
           figwin.canvas.realize()
           figwin.canvas.draw()

   def new_figure_manager(num, *args, **kwargs):
       # in order to expose the Figure constructor to the pylab
       # interface we need to create the figure here
       fig = Figure(*args, **kwargs)
       frame = GameraPlotFrame(num, fig)
       figmgr = frame.get_figure_manager()
       figmgr.canvas.realize()
       figmgr.frame.Show()
       return figmgr

   class GameraFigureManager(backend_bases.FigureManagerBase):
      def __init__(self, canvas, num, frame):
         backend_bases.FigureManagerBase.__init__(self, canvas, num)
         self.frame = frame
         self.window = frame
         self.tb = frame.GetToolBar()
         
         def notify_axes_change(fig):
            'this will be called whenever the current axes is changed'        
            if self.tb != None: self.tb.update()
         self.canvas.figure.add_axobserver(notify_axes_change)
        
      def destroy(self, *args):
         self.frame.Destroy()
         self.canvas.Destroy()        
         import wx
         wx.WakeUpIdle()

   from matplotlib import backends
   backends.show = show
   backends.new_figure_manager = new_figure_manager

   matplotlib_installed = True
   
__all__ = "plot show_figure".split()



#
#
# Copyright (C) 2001-2009 Ichiro Fujinaga, Michael Droettboom,
#                         Karl MacMillan, and Christoph Dalitz
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#  
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from gamera.plugin import *
from gamera.gui import has_gui
from gamera import util
import _projections
from math import pi

class projection_rows(PluginFunction):
    """
    Compute the horizontal projections of an image.  This computes the
    number of pixels in each row.
    """
    self_type = ImageType([ONEBIT])
    return_type = IntVector()
    doc_examples = [(ONEBIT,)]

class projection_cols(PluginFunction):
    """
    Compute the vertical projections of an image.  This computes the
    number of pixels in each column.
    """
    self_type = ImageType([ONEBIT])
    return_type = IntVector()
    doc_examples = [(ONEBIT,)]

class projections(PluginFunction):
    """
    Computes the projections in both the *row*- and *column*-
    directions.  This is returned as a tuple (*rows*, *columns*),
    where each element is an ``IntVector`` of projections.
    (Equivalent to ``(image.projections_rows(),
    image.projections_cols())``).

    If the GUI is being used, the result is displayed in a window:

    .. image:: images/projections.png
    """
    self_type = ImageType([ONEBIT])
    return_type = Class()
    pure_python = 1
    def __call__(image):
        rows = _projections.projection_rows(image)
        cols = _projections.projection_cols(image)
        gui = has_gui.gui
        if gui:
            gui.ShowProjections(rows, cols, image)
        return (rows, cols)
    __call__ = staticmethod(__call__)

class projection_skewed_cols(PluginFunction):
    """
    Computes all vertical projections of an image skewed by a list of
    angles. As in rotate_, angles are measured clockwise and in
    degrees.  Thus a rotate followed by a projection_cols would be
    conceptually the same, albeit considerably slower.

    This function is overloaded to work both with a single angle and a
    list of angles as input. In the first case a single projection
    vector is returned, in the second a list of projections
    vectors. This is explained in the following example:

    .. code:: Python

      # called twice with a single angle as input
      proj1 = img.projection_skewed_cols(0.5)
      proj2 = img.projection_skewed_cols(1.0)
   
      # the same result with one function call
      projlist = img.projection_skewed_cols([0.5,1.0])
      proj1 = projlist[0]
      proj2 = projlist[1]

    Note that unlike rotate_ the image size is not extended. Image
    regions moved outside the original image size are simply clipped,
    which restricts this method to small angles.

    .. _rotate: transformation.html#rotate
    """
    category = "Analysis"
    self_type = ImageType([ONEBIT])
    args = Args([FloatVector("Rotation angles")])
    return_type = Class("nested_list")
    author = "Christoph Dalitz"

    def __call__(self, angles):
        if not util.is_sequence(angles):
            return _projections.projection_skewed_cols(self, [angles])[0]
        else:
            return _projections.projection_skewed_cols(self, angles)
    __call__ = staticmethod(__call__)
    doc_examples = [(ONEBIT, 15.0)]

class projection_skewed_rows(PluginFunction):
    """
    Computes all horizontal projections of an image skewed by a list
    of angles. For more details and an example see
    projection_skewed_cols_.

    Note that unlike rotate_ the image size is not extended. Image
    regions moved outside the original image size are simply clipped,
    which restricts this method to small angles.

    .. _projection_skewed_cols: #projection_skewed_cols
    """
    self_type = ImageType([ONEBIT])
    args = Args([FloatVector("Rotation angles")])
    return_type = Class("nested_list")
    author = "Christoph Dalitz"

    def __call__(self, angles):
        if not util.is_sequence(angles):
            return _projections.projection_skewed_rows(self, [angles])[0]
        else:
            return _projections.projection_skewed_rows(self, angles)
    __call__ = staticmethod(__call__)
    doc_examples = [(ONEBIT, 15.0)]

class rotation_angle_projections(PluginFunction):
    """
    Estimates the rotation angle of a document with the aid of skewed
    projections, as described in section 3.1 of C. Dalitz, G.K. Michalakis,
    C. Pranzas: 'Optical Recognition of Psaltic Byzantine Chant Notation.'
    International Journal of Document Analysis and Recognition 11,
    pp. 143-158 (2008).

    This method works for a wide range of documents (text, music,
    forms), but can become slow for large images. This particular
    implementation can be accelerated by reducing the number of black
    pixels in the image, eg. by scaling it down, only considering a
    fraction of the image or by removing 'uninteresting' pixels.

    Arguments:

    *minangle*, *maxangle* (optional):
      angle interval that is searched for the skew angle;
      default values are -2.5 and +2.5

    *accuracy* (optional):
      error bound for the skew angle estimate;
      default value is zero

    When *accuracy* is set to zero, a default value of
    *180\*0.5/(image.ncols\*pi)* is used, which is only a heuristic
    formula for little changes in the projection profile.

    Return Values:

    *rotation angle*:
      The rotation angle necessary to deskew the image.
      Can be used directly as input to rotate_

    *accuracy*:
      Accuracy of the returned angle.

    .. _rotate: transformation.html#rotate
    """
    category = "Analysis"
    self_type = ImageType([ONEBIT])
    args = Args([Float("minangle", default=-2.5), Float("maxangle", default=2.5), Float("accuracy", default=0.0)])
    return_type = FloatVector("rotation_angle_and_accuracy", 2)
    author = "Christoph Dalitz"
    pure_python = 1

    def __call__(self, minangle = -2.5, maxangle = 2.5, accuracy = 0):

        # l2norm: compute L2-Norm of derivative of vec
        def l2norm(vec):
            var = 0
            for i in range(0,len(vec)-1):
                var += (vec[i] - vec[i+1])**2
            return var

        # some arguments checking
        if (accuracy == 0):
            accuracy = 180 * 0.5 / (self.ncols * pi)
        if (maxangle <= minangle):
            raise RuntimeError("maxangle %f must be greater than minangle %f\n" \
                               % (maxangle, minangle))

        #
        # rough guess where the maximum is
        # necessary because l2norm has many local maxima
        #
        roughacc = 0.5
        if ((maxangle - minangle)/4.0) < roughacc:
            # at least five trial points
            roughacc = (maxangle - minangle) / 4.0
        else:
            roughacc = (maxangle-minangle) / round((maxangle-minangle)/roughacc)
        angle = [(minangle + x*roughacc) for x in \
                    range(0,int(round((maxangle - minangle)/roughacc))+1)]
        alist = [l2norm(x) for x in self.projection_skewed_rows(angle)]
        # find maximum in list
        fb = alist[0]; bi = 0
        for i in range(1,len(alist)):
            if alist[i] > fb:
                fb = alist[i]; bi = i
        b = angle[bi]

        #
        # initialize values for golden section search
        #
        if (bi == 0):
            # maximum on lower iterval end: check neighborhood
            c = b + 1.5*accuracy;
            fc = l2norm(self.projection_skewed_rows(c))
            if (fc > fb) and (c < angle[bi+1]):
                a = b; fa = fb;
                b = c; fb = fc;
                c = angle[bi+1]; fc = alist[bi+1];
            else:
                raise RuntimeError("maximum found on interval end %f\n" \
                                   % angle[bi])
        elif (bi == len(angle)-1):
            # maximum on upper iterval end: check neighborhood
            a = b - 1.5*accuracy;
            fa = l2norm(self.projection_skewed_rows(a))
            if (fa > fb) and (a > angle[bi-1]):
                c = b; fc = fb;
                b = a; fb = fa;
                a = angle[bi-1]; fa = alist[bi-1];
            else:
                raise RuntimeError("maximum found on interval end %f\n" \
                                   % angle[bi])
        else:
            # the normal case: maximum somewhere in the middle
            a = angle[bi-1]; fa = alist[bi-1];
            c = angle[bi+1]; fc = alist[bi+1];

        #
        # fine tuning with golden section search
        # see Press at al: "Numerical Recipes",
        # Cambridge University Press (1986)
        #
        golden = 0.38197  # (3 - sqrt(2)) / 2
        x = 500   # dummy value for recognition of first iteration
        while (c-b > accuracy) or (b-a > accuracy):
            if (x == 500):
                # special case first iteration
                if (fc > fa):
                    x = b + golden * (c - b)
                else:
                    x = b - golden * (b - a)
            else:
                # ordinary situation
                if (c-b > b-a):
                    x = b + golden * (c - b)
                else:
                    x = b - golden * (b - a)
            fx = l2norm(self.projection_skewed_rows(x))
            if (x > b):
                if (fx < fb):
                    c = x; fc = fx
                else:
                    a = b; fa = fb;
                    b = x; fb = fx;
            else:
                if (fx < fb):
                    a = x; fa = fx
                else:
                    c = b; fc = fb;
                    b = x; fb = fx;
        return [b, accuracy]

    __call__ = staticmethod(__call__)


class diagonal_projections(PluginFunction):
    """
    Computes diagonal projections of an image by rotating it
    in 45 degrees, and then calculating the horizontal and 
    vertical projections of the rotated image.

    If the GUI is being used, the result is displayed in a window
    """

    self_type = ImageType([ONEBIT])
    return_type = Class()
    pure_python = 1
    def __call__(image):
        rotated_image = image.rotate(45, None, 1)
        rows = _projections.projection_rows(rotated_image)
        cols = _projections.projection_cols(rotated_image)
        gui = has_gui.gui
        if gui:
            gui.ShowProjections(rows, cols, rotated_image)
        return (rows, cols)
    __call__ = staticmethod(__call__)


class ProjectionsModule(PluginModule):
    cpp_headers=["projections.hpp"]
    category = "Analysis"
    functions = [projection_rows, projection_cols, projections,
                 projection_skewed_rows, projection_skewed_cols,
                 rotation_angle_projections, diagonal_projections]
    author = "Michael Droettboom and Karl MacMillan"
    url = "http://gamera.sourceforge.net/"
module = ProjectionsModule()




# -*- mode: python; indent-tabs-mode: nil; tab-width: 3 -*-
# vim: set tabstop=3 shiftwidth=3 expandtab:
#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom,
#                         and Karl MacMillan
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from inspect import isfunction, ismodule
from os import path
import wx
from gamera.gui import toolbar, gui_util

class RuleEngineRunnerTree(wx.TreeCtrl):
   def __init__(self, toplevel, parent):
      self.toplevel = toplevel
      self.modules = []
      self.undo_history = []
      self.added = []
      self.removed = []
      id = wx.NewId()
      wx.TreeCtrl.__init__(self, parent, id)
      self.root = self.AddRoot("RuleSets")
      self.SetItemHasChildren(self.root, True)
      self.SetPyData(self.root, "")
      wx.EVT_TREE_ITEM_ACTIVATED(self, id, self._OnActivated)
      self.Expand(self.root)

   def _OnKey(self, event):
      pass

   def _OnActivated(self, event):
      from gamera import ruleengine
      item = event.GetItem()
      data = self.GetPyData(item)
      multi_display = self.toplevel.multi_iw.id
      single_display = self.toplevel.single_iw.id
      if isfunction(data):
         engine = ruleengine.RuleEngine([data])
      elif isinstance(data, ruleengine.RuleEngine):
         engine = data
      elif ismodule(data):
         engine = ruleengine.RuleEngine(
           [val for val in data.__dict__.values()
            if isinstance(val, ruleengine.RuleEngine)])
      added = {}
      removed = {}
      try:
         added, removed = engine.perform_rules(multi_display.GetAllItems())
      except ruleengine.RuleEngineError, e:
         gui_util.message(str(e))
         return
      if len(added) == 0 and len(removed) == 0:
         gui_util.message("No rules matched the data set.")
      multi_display.ClearSelection()
      multi_display.append_and_remove_glyphs(added, removed)
      single_display.highlight_cc(added, 0)
      # single_display.add_highlight_cc(removed, 1)
      self.undo_history.append((added, removed))
      self.added, self.removed = added, removed

   def _OnUndo(self, event):
      multi_display = self.toplevel.multi_iw.id
      single_display = self.toplevel.single_iw.id
      if len(self.undo_history):
         added, removed = self.undo_history.pop()
         multi_display.ClearSelection()
         for glyph in removed:
            del glyph.__dict__['dead']
         multi_display.append_and_remove_glyphs([], added)
         if len(self.undo_history):
            self.added, self.removed = self.undo_history[-1]
            single_display.highlight_cc(self.added, 0)
            single_display.add_highlight_cc(self.removed, 1)

   def _OnReloadCode(self, event):
      for module in self.modules:
         reload(module)
      self.CollapseAndReset(self.root)
      old_modules = self.modules[:]
      self.modules = []
      for module in old_modules:
         self.add_module(module)

   def _OnSelectAdded(self, event):
      self.toplevel.multi_iw.id.SelectGlyphs(self.added)

   def _OnSelectRemoved(self, event):
      self.toplevel.multi_iw.id.SelectGlyphs(self.removed)

   def add_module(self, module):
      from gamera import ruleengine
      self.modules.append(module)
      module_node = self.AppendItem(self.root, path.split(module.__file__)[1])
      self.SetPyData(module_node, module)
      for key, val in module.__dict__.items():
         self.SetItemHasChildren(module_node, True)
         if isinstance(val, ruleengine.RuleEngine):
            rule_engine_node = self.AppendItem(module_node, key)
            self.Expand(module_node)
            self.SetPyData(rule_engine_node, val)
            for rule in val.get_rules():
               self.SetItemHasChildren(rule_engine_node, True)
               rule_node = self.AppendItem(rule_engine_node, rule.__name__)
               self.SetPyData(rule_node, rule)

class RuleEngineRunnerPanel(wx.Panel):
   def __init__(self, toplevel, parent = None, id = -1):
      wx.Panel.__init__(
        self, parent, id,
        style=wx.WANTS_CHARS|wx.CLIP_CHILDREN|wx.NO_FULL_REPAINT_ON_RESIZE)
      self.toplevel = toplevel
      self.SetAutoLayout(True)
      self.toolbar = toolbar.ToolBar(self, -1)
      from gamera.gui import gamera_icons
      self.tree = RuleEngineRunnerTree(toplevel, self)
      self.toolbar.AddSimpleTool(
        10, gamera_icons.getIconUndoBitmap(),
        "Undo", self.tree._OnUndo)
      self.toolbar.AddSeparator()
      self.toolbar.AddSimpleTool(
        11, gamera_icons.getIconRefreshBitmap(),
        "Reload code", self.tree._OnReloadCode)
      self.toolbar.AddSeparator()
      self.toolbar.AddSimpleTool(
        12, gamera_icons.getIconSelectAddedBitmap(),
        "Select added glyphs", self.tree._OnSelectAdded)
      self.toolbar.AddSimpleTool(
        13, gamera_icons.getIconSelectRemovedBitmap(),
        "Select removed glyphs", self.tree._OnSelectRemoved)

      lc = wx.LayoutConstraints()
      lc.top.SameAs(self, wx.Top, 0)
      lc.left.SameAs(self, wx.Left, 0)
      lc.right.SameAs(self, wx.Right, 0)
      lc.height.AsIs()
      self.toolbar.SetAutoLayout(True)
      self.toolbar.SetConstraints(lc)
      lc = wx.LayoutConstraints()
      lc.top.Below(self.toolbar, 0)
      lc.left.SameAs(self, wx.Left, 0)
      lc.right.SameAs(self, wx.Right, 0)
      lc.bottom.SameAs(self, wx.Bottom, 0)
      self.tree.SetAutoLayout(True)
      self.tree.SetConstraints(lc)
      self.Layout()

   def open_module(self, filename):
      import imp
      try:
         filename = filename.encode('utf8')
      except Exception:
         pass
      fd = open(filename, 'r')
      module_name = path.split(filename)[1]
      module_name = module_name[:module_name.rfind('.')]
      module = imp.load_module(
        module_name, fd, filename, ('py', 'r', imp.PY_SOURCE))
      self.tree.add_module(module)



# -*- mode: python; indent-tabs-mode: nil; tab-width: 3 -*-
# vim: set tabstop=3 shiftwidth=3 expandtab:
#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom,
#                          and Karl MacMillan
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# TODO: These are fixed values.  We need an intelligent way to vary them.

# This whole approach to fuzziness is syntactically convenient, but maybe
# not very efficient.

FUDGE_AMOUNT = 3
FUDGE_AMOUNT_2 = 6

from gamera.core import Rect, Point, Dim

# This is a factory function that looks like a constructor
def Fudge(o, amount = FUDGE_AMOUNT):
   # For rectangles, just return a new rectangle that is slightly larger
   if isinstance(o, Rect):
      return Rect(Point(int(o.ul_x-amount), int(o.ul_y - amount)), Dim(int(o.ncols + amount * 2), int(o.nrows + amount * 2)))

   # For integers, return one of our "fudge number proxies"
   elif isinstance(o, int):
      return FudgeInt(o, amount)
   elif isinstance(o, float):
      return FudgeFloat(o, amount)
F = Fudge

class FudgeNumber(object):
   def __lt__(self, other):
      return self.below < other
   def __le__(self, other):
      return self.below <= other
   def __eq__(self, other):
      return self.below <= other and self.above >= other
   def __ne__(self, other):
      return other < self.below and other > self.above
   def __gt__(self, other):
      return self.above > other
   def __ge__(self, other):
      return self.above >= other

class FudgeInt(FudgeNumber, int):
   def __init__(self, value, amount=FUDGE_AMOUNT):
      int.__init__(self, value)
      self.below = int(value - amount)
      self.above = int(value + amount)

class FudgeFloat(FudgeNumber, float):
   def __init__(self, value, amount=FUDGE_AMOUNT):
      int.__init__(self, value)
      self.below = float(value - amount)
      self.above = float(value + amount)



#!/usr/bin/env python

from gamera import core, gendoc

if __name__ == "__main__":
   core.init_gamera()
   gendoc.gendoc(classes=[
      ("gamera.core", "ImageInfo",
       'x_resolution y_resolution ncols nrows depth ncolors'),
      ("gamera.core", "ImageData",
       'nrows ncols page_offset_x page_offset_y stride size bytes pixel_type storage_format'),
      ("gamera.core", "RGBPixel",
       'red green blue hue saturation value cie_x cie_y cie_z cie_Lab_L cie_Lab_a cie_Lab_b cyan magenta yellow')
      ])
   



# -*- mode: python; indent-tabs-mode: nil; tab-width: 3 -*-
# vim: set tabstop=3 shiftwidth=3 expandtab:
#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom, Karl MacMillan
#               2009-2010 Christoph Dalitz
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


"""This module contains the core parts of Gamera including all of the image
representation types, the dimension types, abstract mappings of values
to regions on an image (ImageMap), image loading, and feature generation.
 
Everything in this module can be used without the Gui to facilitate
batch processing. It contains the following classes:

ImageBase - common functionality for all of the image types.
Image - images.
SubImage - a view on part of an image.
Cc - a non-rectangular view of part of an image (the view area
     is marked with labels in the image).
MlCc - the same as Cc but with multiple labels

Additionally this module contains the following functions:

load_image - load an image from a file.
image_info - get information about an image file.
display_multi - display a list of images in a grid-like window.
init_gamera - parse the gamera options and load all of the plugins.
"""

# Python standard library
from array import array
from types import *

# import the classification states
try:
   from gameracore import UNCLASSIFIED, AUTOMATIC, HEURISTIC, MANUAL
except ImportError:
   raise ImportError("Couldn't import the core of Gamera.  Are you trying to start the GUI from the root of the Gamera source tree?  This confuses the Python module loading mechanism.")
# import the pixel types
from gameracore import ONEBIT, GREYSCALE, GREY16, RGB, FLOAT, COMPLEX
from enums import ALL, NONIMAGE
# import the storage types
from gameracore import DENSE, RLE
# import some of the basic types
from gameracore import ImageData, Size, Dim, Point, \
     FloatPoint, Rect, Region, RegionMap, ImageInfo, RGBPixel
# import confidence types
from gameracore import CONFIDENCE_DEFAULT, CONFIDENCE_KNNFRACTION, CONFIDENCE_LINEARWEIGHT, CONFIDENCE_INVERSEWEIGHT, CONFIDENCE_NUN, CONFIDENCE_NNDISTANCE, CONFIDENCE_AVGDISTANCE
# import gamera.gameracore for subclassing
import gameracore
from gamera.gui import has_gui

# from gamera.classify import *
import paths, util    # Gamera-specific
from config import config

class SegmentationError(Exception):
   pass

######################################################################

def load_image(filename, compression = DENSE):
   """**load_image** (FileOpen *filename*, Choice *storage_format* = ``DENSE``)

Load an image from the given filename.  At present, TIFF and PNG files are
supported.

*storage_format*
  The type of `storage format`__ to use for the resulting image.

.. __: image_types.html#storage-formats"""
   from gamera import plugin
   import os.path
   methods = plugin.methods_flat_category("File")
   methods = [y for x, y in methods if x.startswith("load") and x != "load_image"]

   if len(methods) == 0:
      raise RuntimeError("There don't seem to be any imported plugins that can load files.  Try running init_gamera(), or explicitly loading the plugins that support file loading, such as tiff_support and png_support.")

   try:
      filename = filename.encode('utf8')
   except Exception:
      pass
   # First, try being smart by loading by extension
   for method in methods:
      for ext in method.exts:
         if os.path.splitext(filename)[1].lower() == ext.lower():
            try:
               image = method.__call__(filename, compression)
            except Exception:
               pass
            else:
               return image

   # Then just try all options
   for method in methods:
      try:
         image = method.__call__(filename, compression)
      except Exception:
         pass
      else:
         return image

   raise IOError("'%s' could not be loaded." % filename)

def save_image(image, filename):
   """**save_image** (Image(ALL) *image*, String *filename*)

Saves an image to a file.  The file type is automatically
determined from the extension.
"""
   from gamera import plugin
   import os.path
   methods = plugin.methods_flat_category("File")
   methods = [y for x, y in methods if x.startswith("save") and x != "save_image"]

   if len(methods) == 0:
      raise RuntimeError("There don't seem to be any imported plugins that can save files.  Try running init_gamera(), or explicitly loading the plugins that support file saving, such as tiff_support and png_support.")

   for method in methods:
      for ext in method.exts:
         if os.path.splitext(filename)[1][1:].lower() == ext.lower():
            method.__call__(image, filename)
            return
              
   # For backward compatibility, fall back to tiff if
   # we can't automatically determine the filetype by
   # the extension
   from gamera.plugins import _tiff_support
   _tiff_support.save_tiff(image, filename)

def nested_list_to_image(l, t=-1):
   from gamera.plugins import image_utilities
   return image_utilities.nested_list_to_image(l, t)

def image_info(filename):
   """ImageInfo **image_info** (FileOpen *filename*)

Retrieve information about an image file without loading it.

The result is an ImageInfo__ object about a given filename. The ImageInfo
object gives information such as the type (color, greyscale, onebit),
the bit-depth, resolution, size, etc.

.. __: gamera.core.ImageInfo.html"""
   import tiff_support, png_support
   try:
      return tiff_support.tiff_info(filename)
   except RuntimeError:
      try:
         return png_support.PNG_info(filename)
      except RuntimeError:
         raise IOError("File is not a PNG or TIFF file")

def display_multi(list):
   """**display_multi** (ImageList *list*)

Displays a list of images in a grid-like window.  This function has
no effect if the `Gamera GUI`__ is not running.

.. __: gui.html"""
   # If it's not a list, we'll get errors, so make it one
   if not util.is_sequence(list):
      list = [list]
   if not len(list):
      raise ValueError("Given list is empty.")
   return has_gui.gui.ShowImages(list)

# Used to cache the list of all features
all_features = None

class ImageBase:
   """Base class for all of the image objects. This class contains
   common functionality for all of the image types including
   the necessary infrastructure to support dynamically added plugins."""
   # Stores the categorized set of methods.  Bears no relationship
   # to __methods__
   methods = {}

   class Properties(dict):
      def __getitem__(self, attr):
         return dict.get(self, attr, None)
      def __getattr__(self, attr):
         return self.__getitem__(attr)
      def __setattr__(self, attr, value):
         return dict.__setitem__(self, attr, value)
   
   def __init__(self):
      self.name = "Untitled"
      self._display = None
      self.last_display = None
      self.properties = self.__class__.Properties()
      # Keep a list of tuples of (feature_name, feature_function) that
      # have already been generated for this Image
      self.feature_functions = [[], 0]

   def __del__(self):
      if self._display:
         self._display.close()

   def pixel_type_name(self):
      """String **pixel_type_name** ()

Returns a string representing the pixel type of the image. Intended
primarily for display (GUI) purposes, since using the integer values
in ``Image.data.pixel_type`` is more efficient.

See `pixel types`_ for more information."""
      return util.get_pixel_type_name(self.data.pixel_type)
   pixel_type_name = property(pixel_type_name, doc=pixel_type_name.__doc__)

   _storage_format_names = {DENSE:  "Dense",
                            RLE:    "RLE"}
   
   def storage_format_name(self):
      """String **storage_format_name** ()

Returns a string representing the storage format of the image.
Intended primarily for display (GUI) purposes, since using the integer
values in ``Image.data.storage_format`` is more efficient.

See `storage formats`_ for more information."""
      return self._storage_format_names[self.data.storage_format]
   storage_format_name = property(storage_format_name, doc=storage_format_name.__doc__)

   def load_image(filename, compression=DENSE):
      """Load an image from the given filename.  At present, TIFF and PNG files are
supported.

*storage_format*
  The type of `storage format`__ to use for the resulting image.

.. __: image_types.html#storage-formats"""
      return load_image(filename, compression)
   load_image = staticmethod(load_image)

   def save_image(self, filename):
      """Saves an image to a file.  The file type is automatically
determined from the extension."""
      return save_image(self, filename)

   def memory_size(self):
      """Int **memory_size** ()

Returns the memory used by the underlying data of the image (in
bytes).  Note that this memory may be shared with other image views on
the same data."""
      return self.data.bytes
   memory_size = property(memory_size, doc=memory_size.__doc__)

   def set_display(self, _display):
      self._display = _display

   def display(self, title=None):
      """Displays the image in its own window.  (See `Using the Gamera GUI`__).  If the GUI
process is not running, this method has no effect.

.. __: gui.html

.. image:: images/display.png"""
      if self.data.pixel_type in (COMPLEX, FLOAT):
         conversion = "to_greyscale"
      else:
         conversion = None
      if self._display:
         self._display.set_image(self, conversion)
      else:
         if title == None:
            self.set_display(
               has_gui.gui.ShowImage(self, self.name, conversion,
                                     owner=self))
         else:
            self.set_display(
               has_gui.gui.ShowImage(self, title, conversion,
                                     owner=self))
      self.last_display = "normal"
      return self._display

   def display_ccs(self):
      """Displays the image in its own window.  (See `Using the Gamera GUI`__).
Each connected component is assigned to one of eight colors.  This
display can be used to see how connected component analysis performs
on a given image.  Uses color_ccs_ under the hood.

.. __: gui.html
.. _color_ccs: gui_support.html#color-ccs

.. note: Connected component analysis must already be performed on the image
   (using cc_analysis_, for example) in order for this to work.

.. _cc_analysis: segmentation.html#cc-analysis

.. image:: images/display_ccs.png"""
      if self._display:
         self._display.set_image(self, "color_ccs")
      else:
         self.set_display(
            has_gui.gui.ShowImage(self, self.name, "color_ccs",
                          owner=self))
      self.last_display = "normal"
      return self._display

   def display_false_color(self):
      """Displays the image using false coloring.  (See false_color_).

.. _false_color: color.html#false-color

.. image:: images/display_false_color.png"""
      if self._display:
         self._display.set_image(self, "false_color")
      else:
         self.set_display(
            has_gui.gui.ShowImage(self, self.name, "false_color",
                          owner=self))
      self.last_display = "normal"
      return self._display

   def unclassify(self):
      """Sets the image back to an UNCLASSIFIED state.  Use this
when you are not longer sure of the identity of the image and you
want an automatic classifier to reclassify."""
      self.id_name = []
      self.confidence = {}
      self.classification_state = UNCLASSIFIED

   def classify_manual(self, id_name):
      """Classifies the image as the value *id_name* and sets the state
to MANUAL.  Use this method when an end user has classified this glyph.
This method also unsets the *confidence* map.

*id_name*
  Can come in one of two forms:

    **string**
       image is classified using the given ``.``-delimited class name.
       As in this case no confidence is provided, it is set to one.
    **list of tuples**
       A list of tuples where each tuple is the pair (*confidence*, *class_name*).

       *confidence*
         A value in range (0, 1), where 0 is uncertain and 1 is certain.
       *class_name*
         A ``.``-delimited class name."""
      if util.is_string_or_unicode(id_name):
         id_name = [(1.0, id_name)]
      elif type(id_name) != ListType:
         raise TypeError("id_name must be a string or a list")
      self.id_name = id_name
      self.confidence = {}
      self.classification_state = MANUAL

   def classify_automatic(self, id_name):
      """Classifies the image as the value *id_name* and sets the state
to AUTOMATIC.  Use this method when an automatic classifier has classified
this glyph. Note that this method does not touch the *confidence* map for
the main id; this must be set separately.

*id_name*
  Can come in one of two forms:

    **string**
       image is classified using the given ``.``-delimited class name.
       As in this case no confidence is provided, it is set to zero.
    **list of tuples**
       A list of tuples where each tuple is the pair (*confidence*, *class_name*).

       *confidence*
         A value in range (0, 1), where 0 is uncertain and 1 is certain.
       *class_name*
         A ``.``-delimited class name."""
      if util.is_string_or_unicode(id_name):
         id_name = [(0.0, id_name)]
      elif type(id_name) != ListType:
         raise TypeError("id_name must be a string or a list")
      self.id_name = id_name
      self.classification_state = AUTOMATIC

   def classify_heuristic(self, id_name):
      """Classifies the image as the value *id_name* and sets the state
to HEURISTIC.  Use this method when a heuristic process has classified
this glyph. This method also unsets the *confidence* map.

*id_name*
  Can come in one of two forms:

    **string**
       image is classified using the given ``.``-delimited class name.
       As in this case no confidence is provided, it is set to 0.5.
    **list of tuples**
       A list of tuples where each tuple is the pair (*confidence*, *class_name*).

       *confidence*
         A value in range (0, 1), where 0 is uncertain and 1 is certain.
       *class_name*
         A ``.``-delimited class name."""
      if util.is_string_or_unicode(id_name):
         id_name = [(0.5, id_name)]
      elif type(id_name) != ListType:
         raise TypeError("id_name must be a string or a list")
      self.id_name = id_name
      self.confidence = {}
      self.classification_state = HEURISTIC

   def get_main_id(self):
      """Returns the primary class name assigned to the image.

If the image is unclassified, the result is 'UNCLASSIFIED'.
"""
      if self.classification_state == UNCLASSIFIED or not len(self.id_name):
         return 'UNCLASSIFIED'
      return self.id_name[0][1]

   def get_confidence(self, confidence_type=None):
      """Returns the confidence of the primary class assigned to the image.
*confidence_type* can be one of the predefined `confidence measures`__, e.g.

.. __: classify.html#confidence

.. code:: Python

   from gamera.gameracore import CONFIDENCE_KNNFRACTION
   glyph.get_confidence(CONFIDENCE_KNNFRACTION)

When no *confidence_type* is given, the default confidence stored with
``id_name[0]`` is returned. If the image is unclassified, -1.0 is returned.

For details about the confidence computation in Gamera, see

   C. Dalitz: `Reject Options and Confidence Measures for kNN Classifiers.`__
   In C. Dalitz (Ed.): "Document Image Analysis with the Gamera Framework."
   Schriftenreihe des Fachbereichs Elektrotechnik und Informatik,
   Hochschule Niederrhein, vol. 8, pp. 16-38, Shaker Verlag (2009).

.. __: http://lionel.kr.hsnr.de/~dalitz/data/publications/sr09-knn-rejection.pdf
"""
      if self.classification_state == UNCLASSIFIED:
         return -1.0
      if not confidence_type:
         return self.id_name[0][0]
      else:
         if self.confidence.has_key(confidence_type):
            return self.confidence[confidence_type]
         else:
            raise ValueError("Given confidence %i not stored in confidence map of image" % confidence_type)

   def has_id_name(self, name):
      """Returns ``True`` if the image has the given classification ``name``."""
      for confidence, id_name in self.id_name:
         if name == id_name:
            return True
      return False

   def subimage(self, *args, **kwargs):
      """Creates a new view on existing data.

There are a number of ways to create a subimage:

   - subimage(Point *upper_left*, Point *lower_right*)
      
   - subimage(Point *upper_left*, Size *size*)

   - subimage(Point *upper_left*, Dim *dim*)

   - subimage(Rect *rectangle*)

Changes to subimages will affect all other subimages viewing the same data.

.. warning:: The *upper_left* and *lower_right* coordinates are absolute and
   not relative to the image origin. Hence, for all practical use cases,
   you must add the image offset to the coordinates, e.g.

   .. code:: Python

      subimg = image.subimage(Point(p.x + image.offset_x, p.y + image.offset_y), dim)
      # or, equivalently:
      subimg = image.subimage(p + image.ul, dim)

"""
      if hasattr(self, "label"):
         return Cc(self, self.label, *args, **kwargs)
      if self.__class__.__name__=="MlCc": #seems more robust to me
         return self.copy(*args, **kwargs)
      else:
         return SubImage(self, *args, **kwargs)

   def _get_feature_vector_size(cls, functions):
      num_features = 0
      for name, function in functions:
          num_features += function.return_type.length
      return num_features
   _get_feature_vector_size = classmethod(_get_feature_vector_size)

   def get_feature_functions(cls, features='all'):
      from gamera import plugin
      global all_features
      if all_features is None:
         all_features = plugin.methods_flat_category('Features', ONEBIT)
         all_features.sort()
      if features == 'all' or features is None:
         functions = all_features
         return functions, cls._get_feature_vector_size(functions)
      features = util.make_sequence(features)
      all_strings = True
      for feature in features:
         if not util.is_string_or_unicode(feature):
            all_strings = False
            break
      if not all_strings:
         import plugin
         all_functions = False
         if (type(features) == tuple and
             len(features) == 2 and
             type(features[0]) == list and
             type(features[1]) == int):
            all_functions = True
            for feature in features[0]:
               if not (type(feature) == tuple and
                       util.is_string_or_unicode(feature[0]) and
                       issubclass(feature[1], plugin.PluginFunction)):
                  all_functions = False
                  break
         if not all_functions:
            raise ValueError(
               "'%s' is not a valid way to specify a list of features."
               % str(features))
         else:
            return features
      else:
         features.sort()
         functions = []
         for feature in features:
            found = 0
            for i in all_features:
               if feature == i[0]:
                  functions.append(i)
                  found = 1
                  break
            if not found:
               raise ValueError("'%s' is not a known feature function." % feature)
         functions.sort()
         return functions, cls._get_feature_vector_size(functions)
   get_feature_functions = classmethod(get_feature_functions)

   def to_xml(self, stream=None):
      """
      Returns a string containing the Gamera XML representation of the image.
      (See the Gamera XML DTD in ``misc/gamera.dtd`` in the source distribution.)
      """
      import gamera_xml
      return gamera_xml.WriteXML(glyphs=[self]).write_stream(stream)

   def to_xml_filename(self, filename):
      """
      Saves the Gamera XML representation of the image to the given *filename*.
      (See the Gamera XML DTD in ``misc/gamera.dtd`` in the source distribution.)
      """
      import gamera_xml
      return gamera_xml.WriteXML(glyphs=[self]).write_filename(filename)

   def set_property(self, name, value):
      self.property[name] = value

   def get_property(self, name):
      return self.property[name]

######################################################################

class Image(gameracore.Image, ImageBase):
   def __init__(self, *args, **kwargs):
      ImageBase.__init__(self)
      gameracore.Image.__init__(self, *args, **kwargs)
   __init__.__doc__ = gameracore.Image.__doc__

   def __del__(self):
      if self._display:
         self._display.close()

######################################################################

class SubImage(gameracore.SubImage, ImageBase):
   def __init__(self, *args, **kwargs):
      ImageBase.__init__(self)
      gameracore.SubImage.__init__(self, *args, **kwargs)
   __init__.__doc__ = gameracore.SubImage.__doc__

   def __del__(self):
      if self._display:
         self._display.close()


######################################################################

class Cc(gameracore.Cc, ImageBase):
   def __init__(self, *args, **kwargs):
      ImageBase.__init__(self)
      gameracore.Cc.__init__(self, *args, **kwargs)
   __init__.__doc__ = gameracore.Cc.__doc__

   def __del__(self):
      if self._display:
         self._display.close()
   
   def display_context(self):
      """**display_context** ()

Displays the Cc in context, by opening a window displaying the underlying
image with the given Cc highlighted.  This method has no effect if the GUI
is not running.
"""
      image = self.parent()
      if self._display:
         self._display.set_image(image)
      else:
         self.set_display(
            has_gui.gui.ShowImage(image,
                          "Connected Component",
                          owner=self))
      self._display.highlight_cc(self)
      self._display.focus(self)
      self.last_display = "context"

class MlCc(gameracore.MlCc, ImageBase):
   def __init__(self, *args, **kwargs):
      ImageBase.__init__(self)
      gameracore.MlCc.__init__(self, *args, **kwargs)
   __init__.__doc__ = gameracore.MlCc.__doc__

   def __del__(self):
      if self._display:
         self._display.close()

# this is a convenience function for using in a console
_gamera_initialised = False
def _init_gamera():
   global _gamera_initialised
   if _gamera_initialised:
      return
   _gamera_initialised = True
   import plugin, gamera_xml, sys
   from gamera.args import NoneDefault
   # Create the default functions for the menupl
   for method in (
      plugin.PluginFactory(
         "load_image", "File", plugin.ImageType(ALL, "image"),
         plugin.ImageType(ALL), plugin.Args([plugin.FileOpen("filename", extension=util.load_image_file_extension_finder)])),
      plugin.PluginFactory(
         "save_image", "File", None,
         plugin.ImageType(ALL), plugin.Args([plugin.FileSave("filename", extension=util.save_image_file_extension_finder)])),
      plugin.PluginFactory(
         "display", "Displaying", None, plugin.ImageType(ALL), None),
      plugin.PluginFactory(
         "display_ccs", "Displaying", None, plugin.ImageType([ONEBIT]),
         None),
      plugin.PluginFactory(
         "display_false_color", "Displaying", None,
         plugin.ImageType([GREYSCALE, FLOAT]),
         None),
      plugin.PluginFactory(
         "classify_manual", "Classification", None,
         plugin.ImageType([ONEBIT]), plugin.Args([plugin.String("id")])),
      plugin.PluginFactory(
         "classify_heuristic", "Classification", None,
         plugin.ImageType([ONEBIT]), plugin.Args([plugin.String("id")])),
      plugin.PluginFactory(
         "classify_automatic", "Classification", None,
         plugin.ImageType([ONEBIT]), plugin.Args([plugin.String("id")])),
      plugin.PluginFactory(
         "unclassify", "Classification", None,
         plugin.ImageType([ONEBIT]), None),
      plugin.PluginFactory(
         "get_main_id", "Classification", plugin.String("id"),
         plugin.ImageType([ONEBIT]), None),
      plugin.PluginFactory(
         "get_confidence", "Classification", plugin.Float("confidence"),
         plugin.ImageType([ONEBIT]), plugin.Args([plugin.Choice("confidence_type", default=NoneDefault)])),
      plugin.PluginFactory(
         "has_id_name", "Classification", plugin.Check("result"),
         plugin.ImageType([ONEBIT]), plugin.Args([plugin.String("id")])),
      plugin.PluginFactory(
         #"subimage", "Utility", plugin.Check("result"),
         "subimage", "Utility", plugin.ImageType(ALL),
         plugin.ImageType(ALL), plugin.Args([plugin.Point("upper_left"),
                                             plugin.Point("lower_right")])),
      plugin.PluginFactory(
         "to_xml", "XML", plugin.String('xml'),
         plugin.ImageType([ONEBIT]), None),
      plugin.PluginFactory(
         "to_xml_filename", "XML", None, plugin.ImageType([ONEBIT]),
         plugin.Args([
      plugin.FileSave("filename", extension=gamera_xml.extensions)]))
      ):
      method.register()
   try:
      verbose = config.get("verbosity_level")
   except Exception:
      verbose = 0
   paths.import_directory(paths.plugins, globals(), locals(), verbose)
   sys.path.append(".")

import sys
if sys.platform == 'win32':
   # Windows doesn't generally keep the console window open, making it difficult to
   # diagnose fatal errors.  This catch-all should help.
   def init_gamera():
      try:
         _init_gamera()
      except Exception, e:
         print type(e)
         if not isinstance(e, SystemExit):
            import traceback
            print "Gamera made a fatal error:"
            print
            traceback.print_exc()
            print
            print "Press <ENTER> to exit."
            x = raw_input()
            sys.exit(1)
else:
   init_gamera = _init_gamera

if __name__ == "__main__":
   init_gamera()

__all__ = ("init_gamera UNCLASSIFIED AUTOMATIC HEURISTIC MANUAL "
           "ONEBIT GREYSCALE GREY16 RGB FLOAT COMPLEX ALL DENSE RLE "
           "CONFIDENCE_DEFAULT CONFIDENCE_KNNFRACTION "
           "CONFIDENCE_LINEARWEIGHT CONFIDENCE_INVERSEWEIGHT "
           "CONFIDENCE_NUN CONFIDENCE_NNDISTANCE CONFIDENCE_AVGDISTANCE "
           "ImageData Size Dim Point FloatPoint Rect Region RegionMap "
           "ImageInfo Image SubImage Cc MlCc load_image image_info "
           "display_multi ImageBase nested_list_to_image RGBPixel "
           "save_image").split()



import py.test

from gamera.core import *
init_gamera()

from gamera import gamera_xml

features = ['aspect_ratio', 'moments', 'volume64regions']

# check for identical training data *except features*
# feature values are subject to rounding errors and
# are thus randomly different => test files not portable
def equal_files(filea, fileb, gz=False):
   if gz:
      import gzip
      a = gzip.open(filea, "r")
      b = gzip.open(fileb, "r")
   else:
      a = open(filea, "r")
      b = open(fileb, "r")
   i = 0
   linea = a.readline()
   lineb = b.readline()
   feature = False
   while linea:
      i += 1
      if not lineb:
         # a longer than b
         return False
      if linea != lineb:
         return False
      if "<feature name=" in linea:
         while linea and "</feature>" not in linea:
            linea = a.readline()
         while lineb and "</feature>" not in lineb:
            lineb = b.readline()
      linea = a.readline()
      lineb = b.readline()
   if lineb:
      # b longer than a
      return False
   return True

# High-level API
def test_glyphs_from_xml():
   glyphs = gamera_xml.glyphs_from_xml("data/testline.xml")
   assert len(glyphs) == 66
      
def test_glyphs_with_features_from_xml():
   glyphs = gamera_xml.glyphs_with_features_from_xml(
      "data/testline.xml", ["area", "aspect_ratio"])
   assert len(glyphs) == 66
   assert len(glyphs[0].features) == 2

def test_glyphs_to_xml():
   glyphs = gamera_xml.glyphs_from_xml("data/testline.xml")
   gamera_xml.glyphs_to_xml("tmp/testline_test1.xml", glyphs, False)
   assert equal_files("tmp/testline_test1.xml", "data/testline_test1.xml")

def test_glyphs_to_xml_with_features():
   glyphs = gamera_xml.glyphs_with_features_from_xml("data/testline.xml", feature_functions=features)
   gamera_xml.glyphs_to_xml("tmp/testline_test2.xml", glyphs, True)
   assert equal_files("tmp/testline_test2.xml", "data/testline_test2.xml")

def test_glyphs_from_xml_gz():
   glyphs = gamera_xml.glyphs_from_xml("data/testline.xml.gz")
   assert len(glyphs) == 66

def test_glyphs_to_xml_gz():
   glyphs = gamera_xml.glyphs_from_xml("data/testline.xml")
   gamera_xml.glyphs_to_xml("tmp/testline_test1.xml.gz", glyphs, False)
   assert equal_files("tmp/testline_test1.xml.gz", "data/testline_test1.xml.gz", gz=True)

# Low-level API
def test_write_xml():
   glyphs = gamera_xml.glyphs_from_xml("data/testline.xml")
   writer = gamera_xml.WriteXML(glyphs)
   result_string = writer.string()
   writer.write_filename("tmp/testline_test3.xml")
   assert equal_files("tmp/testline_test3.xml", "data/testline_test3.xml")

def test_symbol_table():
   symbol_table = gamera_xml.LoadXML(parts=['symbol_table']).parse_filename("data/symbol_table.xml").symbol_table
   gamera_xml.WriteXMLFile([], symbol_table).write_filename("tmp/symbol_table.xml")
   assert equal_files("data/symbol_table.xml", "tmp/symbol_table.xml")

   symbol_table = gamera_xml.LoadXML(parts=['symbol_table']).parse_filename("data/testline.xml").symbol_table
   gamera_xml.WriteXMLFile([], symbol_table).write_filename("tmp/symbol_table2.xml")
   assert equal_files("data/symbol_table.xml", "tmp/symbol_table2.xml")

# Error cases
def test_missing_attributes():
   def _test_missing_attributes():
      glyphs = gamera_xml.glyphs_from_xml("data/missing_attributes.xml")
   py.test.raises(gamera_xml.XMLError, _test_missing_attributes)

def test_malformed():
   def _test_malformed():
      glyphs = gamera_xml.glyphs_from_xml("data/malformed.xml")
   py.test.raises(gamera_xml.XMLError, _test_malformed)



#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom, and Karl MacMillan
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#  
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Various functions related to corelation (template matching)."""

from gamera.plugin import *
from gamera import util
import _corelation

class corelation_weighted(PluginFunction):
    """
    Returns a floating-point value for how well an image is corelated
    to another image placed at a given origin (*x*, *y*).  Uses the
    weighted reward/penalty method.

    *template*
      The template image.

    *offset*
      The displacement of the template on the image.

    *bb*, *bw*, *wb*, *ww*
      The rewards and penalties for different combinations of pixels.
      The first letter in the argument name indicates the color of the
      template; the second letter indicates the color of the source
      image.  For instance, the value of *bw* will be applied to the
      result when the template pixel is black and the source image
      pixel is white.

    +--------+--------+------------------+
    | Image  |        | Template         |
    |        +--------+---------+--------+
    |        |        | black   | white  |
    |        +--------+---------+--------+
    |        | black  | *bb*    | *wb*   |
    |        +--------+---------+--------+
    |        | white  | *bw*    | *ww*   |
    +--------+--------+---------+--------+
    """
    return_type = Float("corelation")
    self_type = ImageType([ONEBIT, GREYSCALE])
    args = Args([ImageType([ONEBIT], "template"),
                 Point("offset"),
                 Float("bb"), Float("bw"), Float("wb"), Float("ww")])

class corelation_sum(PluginFunction):
    """
    Returns a floating-point value for how well an image is corelated
    to another image placed at a given origin (*x*, *y*).  Uses the
    sum of absolute distance method.  A higher value indicates more
    corelation.

    *template*
      The template image.

    *offset*
      The displacement of the template on the image.
    """
    return_type = Float("corelation")
    self_type = ImageType([ONEBIT, GREYSCALE])
    args = Args([ImageType([ONEBIT], "template"), Point("offset")])
    progress_bar = "Correlating"

class corelation_sum_squares(PluginFunction):
    """
    Returns a floating-point value for how well an image is corelated
    to another image placed at a given origin (*x*, *y*).  Uses the
    sum of squares method.  A higher value indicates more corelation.

    *template*
      The template image.
    *offset*
      The displacement of the template on the image.
    """
    return_type = Float("corelation")
    self_type = ImageType([ONEBIT, GREYSCALE])
    args = Args([ImageType([ONEBIT], "template"), Point("offset")])
    progress_bar = "Correlating"

class CorelationModule(PluginModule):
    cpp_headers=["corelation.hpp"]
    category = "Corelation"
    functions = [corelation_weighted, corelation_sum,
                 corelation_sum_squares]
    author = "Michael Droettboom"
    url = "http://gamera.sourceforge.net/"
module = CorelationModule()



# -*- mode: python; indent-tabs-mode: nil; tab-width: 3 -*-
# vim: set tabstop=3 shiftwidth=3 expandtab:
#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom,
#                         and Karl MacMillan
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

(NO_GUI, WX_GUI, CURSES_GUI) = tuple(range(3))
has_gui = NO_GUI

class NullGui:
   class NullMethod:
      def __call__(*args, **kwargs):
         pass
   
   def __getattr__(self, attr):
      return self.NullMethod()
   
gui = NullGui()
